"""
Functionality of this module that extends Numerical Python
==========================================================

The following extensions to Numerical Python are also defined:

 - seq
           seq(a,b,s, [type]) computes numbers from a up to and
           including b in steps of s and (default) type float_
           sequence = seq (for backward compatibility)

 - iseq:
           as seq, but integer counters are computed
           (iseq is an alternative to range where the
           upper limit is included in the sequence - this can
           be important for direct mapping of indices between
           mathematics and Python code)
           isequence = iseq (for backward compatibility)

 - arr:
           simplified/unified interface to creating NumPy
           arrays (see its doc string)

 - solve_tridiag_linear_system:
           returns the solution of a tridiagonal linear system

 - wrap2callable:
           tool for turning constants, discrete data, string
           formulas, function objects, or plain functions
           into an object that behaves as a function

 - NumPy_array_iterator:
           allows iterating over all array elements using
           a single, standard for loop (for value, index in iterator),
           has some additional features compared with numpy.ndenumerate
             
 - asarray_cpwarn:
           as asarray(a), but a warning or exception is issued if
           the array a is copied

 - ndgrid:
           extend one-dimensional coordinate arrays to multi-dimensional
           arrays over grids
           
 - float_eq:
           operator == for float operands with tolerance,
           float_eq(a,b,tol) means abs(a-b) < tol
           works for both scalar and array arguments

 - norm_L2, norm_l2, norm_L1, norm_l1, norm_inf: 
           norms for multi-dimensional arrays viewed as vectors

 - compute_historgram:
           return x and y arrays of a histogram, given a vector of samples

 - factorial:
           compute the factorial n! by various methods (iterative,
           recursive, reduce, functional, scipy, etc)

"""

if __name__.find('numpyutils') != -1:
    from numpy import *

#else if name is some other module name:
# this file is included in numpytools.py (through a preprocessing step)
# and the code below then relies on previously imported Numerical Python
# modules (Numeric, numpy, numarray)

import operator

def asarray_cpwarn(a, dtype=None, message='warning', comment=''):
    """
    As asarray, but a warning or exception is issued if the
    a array is copied.
    """
    a_new = asarray(a, dtype)
    # must drop numpy's order argument since it conflicts
    # with Numeric's savespace

    # did we copy?
    if a_new is not a:
        # we do not return the identical array, i.e., copy has taken place
        msg = '%s  copy of array %s, from %s to %s' % \
              (comment, a.shape, type(a), type(a_new))
        if message == 'warning':
            print 'Warning: %s' % msg
        elif message == 'exception':
            raise TypeError, msg
    return a_new


def seq(min=0.0, max=None, inc=1.0, type=float,
        return_type='NumPyArray'):
    """
    Generate numbers from min to (and including!) max,
    with increment of inc. Safe alternative to arange.
    The return_type string governs the type of the returned
    sequence of numbers ('NumPyArray', 'list', or 'tuple').
    """
    if max is None: # allow sequence(3) to be 0., 1., 2., 3.
        # take 1st arg as max, min as 0, and inc=1
        max = min; min = 0.0; inc = 1.0
    r = arange(min, max + inc/2.0, inc, type)
    if return_type == 'NumPyArray' or return_type == ndarray:
        return r
    elif return_type == 'list':
        return r.tolist()
    elif return_type == 'tuple':
        return tuple(r.tolist())


def iseq(start=0, stop=None, inc=1):
    """
    Generate integers from start to (and including) stop,
    with increment of inc. Alternative to range/xrange.
    """
    if stop is None: # allow isequence(3) to be 0, 1, 2, 3
        # take 1st arg as stop, start as 0, and inc=1
        stop = start; start = 0; inc = 1
    return xrange(start, stop+inc, inc)

sequence = seq  # backward compatibility
isequence = iseq  # backward compatibility


def arr(shape=None, element_type=float,
        interval=None, 
        data=None, copy=True,
        file_=None,
        order='C'):
    """
    Compact and flexible interface for creating NumPy arrays,
    including several consistency and error checks.

    @param shape:        length of each dimension
    @type  shape:        tuple or int
    @param data:         list, tuple, or NumPy array with data elements
    @param copy:         copy data if true, share data if false
    @type  copy:         boolean
    @param element_type: float, int, int16, float64, bool, etc.
    @param interval:     make elements from a to b (shape gives no of elms)
    @type  interval:     tuple or list
    @param file_:        filename or file object containing array data
    @type  file_:        string
    @param order:        'Fortran' or 'C' storage
    @type  order:        string
    @return:             created Numerical Python array

    The array can be created in four ways:
    
      1. as zeros (just shape specified),

      2. as uniformly spaced coordinates in an interval [a,b]

      3. as a copy of or reference to (depending on copy=True,False resp.)
         a list, tuple, or NumPy array (provided as the data argument),

      4. from data in a file (for one- or two-dimensional real-valued arrays).

    The function calls the underlying NumPy functions zeros, array and
    linspace (see the NumPy manual for the functionality of these
    functions).  In case of data in a file, the first line determines
    the number of columns in the array. The file format is just rows
    and columns with numbers, no decorations (square brackets, commas,
    etc.) are allowed.

    >>> arr((3,4))
    array([[ 0.,  0.,  0.,  0.],
           [ 0.,  0.,  0.,  0.],
           [ 0.,  0.,  0.,  0.]])

    >>> arr(4, element_type=int) + 4  # integer array
    array([4, 4, 4, 4])

    >>> arr(3, interval=[0,2])
    array([ 0.,  1.,  2.])
           
    >>> somelist=[[0,1],[5,5]]
    >>> a = arr(data=somelist)
    >>> a  # a has always float elements by default
    array([[ 0.,  1.],
           [ 5.,  5.]])
    >>> a = arr(data=somelist, element_type=int)
    >>> a
    array([[0, 1],
           [5, 5]])
    >>> b = a + 1
    
    >>> c = arr(data=b, copy=False)  # let c share data with b
    >>> b is c
    True
    >>> id(b) == id(c)
    True

    >>> # make a file with array data:
    >>> f = open('tmp.dat', 'w')
    >>> f.write('''\
    ... 1 3
    ... 2 6
    ... 3 12
    ... 3.5 20
    ... ''')
    >>> f.close()
    >>> # read array data from file:
    >>> a = arr(file_='tmp.dat')
    >>> a
    array([[  1. ,   3. ],
           [  2. ,   6. ],
           [  3. ,  12. ],
           [  3.5,  20. ]])
    """
    if data is None and file_ is None and shape is None:
        return None
    
    if data is not None:

        if not operator.isSequenceType(data):
            raise TypeError, 'arr: data argument is not a sequence type'
        
        if isinstance(shape, (list,tuple)):
            # check that shape and data are compatible:
            if reduce(operator.mul, shape) != size(data):
                raise ValueError, \
                      'arr: shape=%s is not compatible with %d '\
                      'elements in the provided data' % (shape, size(data))
        elif isinstance(shape, int):
            if shape != size(data):
                raise ValueError, \
                      'arr: shape=%d is not compatible with %d '\
                      'elements in the provided data' % (shape, size(data))
        elif shape is None:
            if isinstance(data, (list,tuple)) and copy == False:
                # cannot share data (data is list/tuple)
                copy = True
            return array(data, dtype=element_type, copy=copy, order=order)
        else:
            raise TypeError, \
                  'shape is %s, must be list/tuple or int' % type(shape)
    elif file_ is not None:
        if not isinstance(file_, (basestring, file, StringIO)):
            raise TypeError, \
                  'file_ argument must be a string (filename) or '\
                  'open file object, not %s' % type(file_)

        if isinstance(file_, basestring):
            file_ = open(file_, 'r')
        # skip blank lines:
        while True:
            line1 = file_.readline().strip()
            if line1 != '':
                break
        ncolumns = len(line1.split())
        file_.seek(0)
        # we assume that array data in file has element_type=float:
        if not (element_type == float or element_type == 'd'):
            raise ValueError, 'element_type must be float_/"%s", not "%s"' % \
                  ('d', element_type)
        
        d = array([float(word) for word in file_.read().split()])
        if isinstance(file_, basestring):
            f.close()
        # shape array d:
        if ncolumns > 1:
            suggested_shape = (int(len(d)/ncolumns), ncolumns)
            total_size = suggested_shape[0]*suggested_shape[1]
            if total_size != len(d):
                raise ValueError, \
                'found %d array entries in file "%s", but first line\n'\
                'contains %d elements - no shape is compatible with\n'\
                'these values' % (len(d), file, ncolumns)
            d.shape = suggested_shape
        if shape is not None:
            if shape != d.shape:
                raise ValueError, \
                'shape=%s is not compatible with shape %s found in "%s"' % \
                (shape, d.shape, file)
        return d

    elif interval is not None and shape is not None:
        if not isinstance(shape, int):
            raise TypeError, 'For array values in an interval, '\
                  'shape must be an integer'
        if not isinstance(interval, (list,tuple)):
            raise TypeError, 'interval must be list or tuple, not %s' % \
                  type(interval)
        if len(interval) != 2:
            raise ValueError, 'interval must be a 2-tuple (or list)'

        try:
            return linspace(interval[0], interval[1], shape)
        except MemoryError, e:
            # print more information (size of data):
            print e, 'of size %s' % shape

    else:
        # no data, no file, just make zeros

        if not isinstance(shape, (tuple, int, list)):
            raise TypeError, \
           'arr: shape (1st arg) must be tuple or int'
        if shape is None:
            raise ValueError, \
            'arr: either shape, data, or from_function must be specified'

        try:
            return zeros(shape, dtype=element_type, order=order)
        except MemoryError, e:
            # print more information (size of data):
            print e, 'of size %s' % shape
    
def meshgrid(x=None, y=None, z=None, sparse=True, indexing='xy',
             memoryorder=None):
    """
    Make 1D/2D/3D coordinate arrays for vectorized evaluations of
    1D/2D/3D scalar/vector fields over 1D/2D/3D grids, given
    one-dimensional coordinate arrays x, y, and/or, z.

    >>> x=linspace(0,1,3)        # coordinates along x axis
    >>> y=linspace(0,1,2)        # coordinates along y axis
    >>> xv, yv = meshgrid(x,y)   # extend x and y for a 2D xy grid
    >>> xv
    array([[ 0. ,  0.5,  1. ]]
    >>> yv
    array([[ 0.],
           [ 1.]])
    >>> z=5
    >>> xv, yv, zc = meshgrid(x,y,z)  # 2D slice of a 3D grid, with z=const
    >>> xv
    array([[ 0. ,  0.5,  1. ]]
    >>> yv
    array([[ 0.],
           [ 1.]])
    >>> zc
    5

    >>> meshgrid(2,y,x)  # 2D slice of a 3D grid, with x=const
    (2, array([[ 0.,  1.]]), array([[ 0. ],
           [ 0.5],
           [ 1. ]]))
    >>> meshgrid(0,1,5)  # just a 3D point
    (0, 1, 5)
    >>> meshgrid(3)
    3
    >>> meshgrid(y)      # 1D grid; y is just returned
    array([ 0.,  1.])
    >>> meshgrid(x,y,sparse=False)  # store the full 2D matrix
    (array([[ 0. ,  0.5,  1. ],
       [ 0. ,  0.5,  1. ]]), array([[ 0.,  0.,  0.],
       [ 1.,  1.,  1.]]))
    >>> meshgrid(x,y,indexing='ij')  # change to matrix indexing
    (array([[ 0. ],
           [ 0.5],
           [ 1. ]]), array([[ 0.,  1.]]))
    >>> meshgrid(x,y,sparse=False,indexing='ij')
    (array([[ 0. ,  0. ],
           [ 0.5,  0.5],
           [ 1. ,  1. ]]), array([[ 0.,  1.],
           [ 0.,  1.],
           [ 0.,  1.]]))


    Why has SciTools its own meshgrid function when NumPy has three
    similar functions, `mgrid`, `ogrid`, and `meshgrid`?
    The `meshgrid` function in NumPy is limited to two dimensions only, while
    the SciTools version can also work with 3D grids. In addition,
    the NumPy version of `meshgrid` has no option for generating sparse
    grids to conserve memory, like we have in SciTools by specifying the
    `sparse` argument:
    !bc
    >>> xv, yv = meshgrid(linspace(-2,2,5), linspace(-1,1,3), sparse=True)
    >>> xv
    array([[-2., -1.,  0.,  1.,  2.]])
    >>> yv
    array([[-1.],
           [ 0.],
           [ 1.]])
    >>>
    !ec
    Actually, this is the default behavior for the `meshgrid` function in
    SciTools. In NumPy, however, we will in this case get a full 2D grid:
    !bc
    >>> xv, yv = numpy.meshgrid(linspace(-2,2,5), linspace(-1,1,3))
    >>> xv
    array([[-2., -1.,  0.,  1.,  2.],
           [-2., -1.,  0.,  1.,  2.],
           [-2., -1.,  0.,  1.,  2.]])
    >>> yv
    array([[-1., -1., -1., -1., -1.],
           [ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  1.,  1.,  1.,  1.]])
    >>> 
    !ec
    This is the same result we get by setting `sparse=False` in `meshgrid`
    in SciTools.

    The NumPy functions `mgrid` and `ogrid` does provide support for,
    respectively, full and sparse n-dimensional meshgrids, however,
    these functions uses slices to generate the meshgrids rather than
    one-dimensional coordinate arrays such as in Matlab. With slices, the
    user does not have the option to generate meshgrid with, e.g.,
    irregular spacings, like 
    !bc
    >>> x = array([-1,-0.5,1,4,5], float)
    >>> y = array([0,-2,-5], float)
    >>> xv, yv = meshgrid(x, y, sparse=False)
    >>> xv 
    array([[-1. , -0.5,  1. ,  4. ,  5. ],
           [-1. , -0.5,  1. ,  4. ,  5. ],
           [-1. , -0.5,  1. ,  4. ,  5. ]])
    >>> yv
    array([[ 0.,  0.,  0.,  0.,  0.],
           [-2., -2., -2., -2., -2.],
           [-5., -5., -5., -5., -5.]])
    >>> 
    !ec

    In addition to the reasons mentioned above, the meshgrid function in
    NumPy supports only Cartesian indexing, i.e., x and y, not matrix
    indexing, i.e., rows and columns (`mgrid` and `ogrid` supports only
    matrix indexing). The `meshgrid` function in SciTools supports both
    indexing conventions through the `indexing` keyword argument. Giving
    the string `'ij'` returns a meshgrid with matrix indexing, while
    `'xy'` returns a meshgrid with Cartesian indexing. The difference is
    illustrated by the following code snippet:
    !bc
    nx = 10
    ny = 15

    x = linspace(-2,2,nx)
    y = linspace(-2,2,ny)

    xv, yv = meshgrid(x, y, sparse=False, indexing='ij')
    for i in range(nx):
        for j in range(ny):
            # treat xv[i,j], yv[i,j]

    xv, yv = meshgrid(x, y, sparse=False, indexing='xy')
    for i in range(nx):
        for j in range(ny):
            # treat xv[j,i], yv[j,i]
    !ec

    It is not entirely true that matrix indexing is not supported by the
    `meshgrid` function in NumPy because we can just switch the order of
    the first two input and output arguments:
    !bc
    yv, xv = numpy.meshgrid(y, x)
    !ec
    is the same as
    !bc
    xv, yv = meshgrid(x, y, sparse=False, indexing='ij')
    !ec
    However, we think it is clearer to have the logical "x, y"
    sequence on the left-hand side and instead adjust a keyword argument.
    """

    import types
    def fixed(coor):
        return isinstance(coor, (float, complex, int, types.NoneType))

    if not fixed(x):
        x = asarray(x)
    if not fixed(y):
        y = asarray(y)
    if not fixed(z):
        z = asarray(z)
    
    def arr1D(coor):
        try:
            if len(coor.shape) == 1:
                return True
            else:
                return False
        except AttributeError:
            return False
    
    # if two of the arguments are fixed, we have a 1D grid, and
    # the third argument can be reused as is:

    if arr1D(x) and fixed(y) and fixed(z):
        return x
    if fixed(x) and arr1D(y) and fixed(z):
        return y
    if fixed(x) and fixed(y) and arr1D(z):
        return z

    # if x,y,z are identical, make copies:
    try:
        if y is x: y = x.copy()
        if z is x: z = x.copy()
        if z is y: z = y.copy()
    except AttributeError:  # x, y, or z not NumPy array
        pass

    if memoryorder is not None:
        import warnings
        msg = "Keyword argument 'memoryorder' is deprecated and will be " \
              "removed in the future. Please use the 'indexing' keyword " \
              "argument instead."
        warnings.warn(msg, DeprecationWarning)
        if memoryorder == 'xyz':
            indexing = 'ij'
        else:
            indexing = 'xy'
    
    # If the keyword argument sparse is set to False, the full N-D matrix
    # (not only the 1-D vector) should be returned. The mult_fact variable
    # should then be updated as necessary.
    mult_fact = 1

    # if only one argument is fixed, we have a 2D grid:
    if arr1D(x) and arr1D(y) and fixed(z):
        if indexing == 'ij':
            if not sparse:
                mult_fact = ones((len(x),len(y)))
            if z is None:
                return x[:,newaxis]*mult_fact, y[newaxis,:]*mult_fact
            else:
                return x[:,newaxis]*mult_fact, y[newaxis,:]*mult_fact, z
        else:
            if not sparse:
                mult_fact = ones((len(y),len(x)))
            if z is None:
                return x[newaxis,:]*mult_fact, y[:,newaxis]*mult_fact
            else:
                return x[newaxis,:]*mult_fact, y[:,newaxis]*mult_fact, z
        
    if arr1D(x) and fixed(y) and arr1D(z):
        if indexing == 'ij':
            if not sparse:
                mult_fact = ones((len(x),len(z)))
            if y is None:
                return x[:,newaxis]*mult_fact, z[newaxis,:]*mult_fact
            else:
                return x[:,newaxis]*mult_fact, y, z[newaxis,:]*mult_fact
        else:
            if not sparse:
                mult_fact = ones((len(z),len(x)))
            if y is None:
                return x[newaxis,:]*mult_fact, z[:,newaxis]*mult_fact
            else:
                return x[newaxis,:]*mult_fact, y, z[:,newaxis]*mult_fact
        
    if fixed(x) and arr1D(y) and arr1D(z):
        if indexing == 'ij':
            if not sparse:
                mult_fact = ones((len(y),len(z)))
            if x is None:
                return y[:,newaxis]*mult_fact, z[newaxis,:]*mult_fact
            else:
                return x, y[:,newaxis]*mult_fact, z[newaxis,:]*mult_fact
        else:
            if not sparse:
                mult_fact = ones((len(z),len(y)))
            if x is None:
                return y[newaxis,:]*mult_fact, z[:,newaxis]*mult_fact
            else:
                return x, y[newaxis,:]*mult_fact, z[:,newaxis]*mult_fact

    # or maybe we have a full 3D grid:
    if arr1D(x) and arr1D(y) and arr1D(z):
        if indexing == 'ij':
            if not sparse:
                mult_fact = ones((len(x),len(y),len(z)))
            return x[:,newaxis,newaxis]*mult_fact, \
                   y[newaxis,:,newaxis]*mult_fact, \
                   z[newaxis,newaxis,:]*mult_fact
        else:
            if not sparse:
                mult_fact = ones((len(y),len(x),len(z)))
            return x[newaxis,:,newaxis]*mult_fact, \
                   y[:,newaxis,newaxis]*mult_fact, \
                   z[newaxis,newaxis,:]*mult_fact

    # at this stage we assume that we just have scalars:
    l = []
    if x is not None:
        l.append(x)
    if y is not None:
        l.append(y)
    if z is not None:
        l.append(z)
    if len(l) == 1:
        return l[0]
    else:
        return tuple(l)


def ndgrid(*args,**kwargs):
    """
    Same as calling meshgrid with indexing='ij' (see meshgrid for
    documentation).
    """
    kwargs['indexing'] = 'ij'
    return meshgrid(*args,**kwargs)


def float_eq(a, b, rtol=1.0e-14, atol=1.0e-14):
    """
    Approximate test a==b for float variables.
    Returns true if abs(a-b) < atol + rtol*abs(b).
    atol comes into play when abs(b) is large.
    When a and b are NumPy arrays, NumPy's allclose function is called
    (but float_eq's default tolerances are much stricter than those of
    allclose).
    """
    if isinstance(a, float):
        return math.fabs(a-b) < atol + rtol*math.fabs(b)
    elif isinstance(a, complex):
        return float_eq(a.real, b.real, rtol, atol) and \
               float_eq(a.imag, b.imag, rtol, atol)
    else: # assume NumPy array
        try:
            return allclose(a, b, rtol, atol)
        except:
            raise TypeError, 'Illegal types: a is %s and b is %s' % \
                  (type(a), type(b))
    

def norm_l2(u):
    """
    l2 norm of a multi-dimensional array u viewed as a vector
    (norm=sqrt(dot(u.flat,u.flat))).
    """
    return math.sqrt(innerproduct(u.flat, u.flat))

def norm_L2(u):
    """
    L2 norm of a multi-dimensional array u viewed as a vector
    (norm=sqrt(dot(u.flat,u.flat)/n)).

    If u holds function values and the norm of u is supposed to
    approximate an integral (L2 norm) of the function, this (and
    not norm_l2) is the right norm function to use.
    """
    return norm_l2(u)/sqrt(float(size(u)))

def norm_l1(u):
    """
    l1 norm of a multi-dimensional array u viewed as a vector
    (norm=sum(abs(u.flat))).
    """
    return sum(abs(u.flat))

def norm_L1(u):
    """
    L1 norm of a multi-dimensional array u viewed as a vector
    (norm=sum(abs(u.flat))).

    If u holds function values and the norm of u is supposed to
    approximate an integral (L1 norm) of the function, this (and
    not norm_l1) is the right norm function to use.
    """
    return norm_l1(u)/float(size(u))

def norm_inf(u):
    """Infinity/max norm of a multi-dimensional array u viewed as a vector."""
    return abs(u).max()


def solve_tridiag_linear_system(A, b):
    """
    Solve a tridiagonal linear system of the form::
    
     A[0,1]*x[0] + A[0,2]*x[1]                                        = 0
     A[1,0]*x[0] + A[1,1]*x[1] + A[1,2]*x[2]                          = 0
     ...
     ...
              A[k,0]*x[k-1] + A[k,1]*x[k] + A[k,2]*x[k+1]             = 0
     ...
                  A[n-2,0]*x[n-3] + A[n-2,1]*x[n-2] + A[n-2,2]*x[n-1] = 0
     ...
                                    A[n-1,0]*x[n-2] + A[n-1,1]*x[n-1] = 0

    That is, the diagonal is stored in A[:,1], the subdiagonal
    is stored in A[1:,0], and the superdiagonal is stored in A[:n-2,2].
    """

    #The storage is not memory friendly in Python/C (diagonals stored
    #columnwise in A), but if A is sent to F77 for high-performance
    #computing, a copy is taken and the F77 routine works with the
    #same algorithm and hence optimal (columnwise traversal)
    #Fortran storage.
    
    n = len(b)
    x = zeros(n, 'd')  # solution
    # scratch arrays:
    d = zeros(n, 'd');  c = zeros(n, 'd');  m = zeros(n, 'd')

    d[0] = A[0,1]
    c[0] = b[0]

    for k in iseq(start=1, stop=n-1, inc=1):
        m[k] = A[k,0]/d[k-1]
        d[k] = A[k,1] - m[k]*A[k-1,2]
        c[k] = b[k] - m[k]*c[k-1]
    x[n-1] = c[n-1]/d[n-1]

    # back substitution:
    for k in iseq(start=n-2, stop=0, inc=-1):
        x[k] = (c[k] - A[k,2]*x[k+1])/d[k]

    return x




try:
    import Pmw
    class NumPy2BltVector(Pmw.Blt.Vector):
        """
        Copy a NumPy array to a BLT vector:
        # a: some NumPy array
        b = NumPy2BltVector(a)  # b is BLT vector
        g = Pmw.Blt.Graph(someframe)
        # send b to g for plotting
        """
        def __init__(self, array):
            Pmw.Blt.Vector.__init__(self, len(array))
            self.set(tuple(array))  # copy elements
except:
    class NumPy2BltVector:
        def __init__(self, array):
            raise ImportError, "Python is not compiled with BLT"

try:
    from scitools.StringFunction import StringFunction
except:
    pass  # wrap2callable may not work


class WrapNo2Callable:
    """Turn a number (constant) into a callable function."""
    def __init__(self, constant):
        self.constant = constant
        self._array_shape = None

    def __call__(self, *args):
        """
        >>> w = WrapNo2Callable(4.4)
        >>> w(99)
        4.4000000000000004
        >>> # try vectorized computations:
        >>> x = linspace(1, 4, 4)
        >>> y = linspace(1, 2, 2)
        >>> xv = x[:,NewAxis]; yv = y[NewAxis,:]
        >>> xv + yv
        array([[ 2.,  3.],
               [ 3.,  4.],
               [ 4.,  5.],
               [ 5.,  6.]])
        >>> w(xv, yv)
        array([[ 4.4,  4.4],
               [ 4.4,  4.4],
               [ 4.4,  4.4],
               [ 4.4,  4.4]])

        If you want to call such a function object with space-time
        arguments and vectorized expressions, make sure the time
        argument is not the first argument. That is,
        w(xv, yv, t) is fine, but w(t, xv, yv) will return 4.4,
        not the desired array!
        """
        if isinstance(args[0], (float, int, complex)):
            # scalar version:
            # (operator.isNumberType(args[0]) cannot be used as it is
            # true also for NumPy arrays
            return self.constant
        else: # assume NumPy array
            if self._array_shape is None:
                self._set_array_shape()
            else:
                r = self.constant*ones(self._array_shape, 'd')
                # could store r (allocated once) and just return reference
                return r

    def _set_array_shape(self, arg):
        # vectorized version:
        r = arg.copy()
        # to get right dimension of the return array,
        # compute with args in a simple formula (sum of args)
        for a in args[1:]:
            r = r + a  # in-place r+= won't work
            # (handles x,y,t - the last t just adds a constant)
            # an argument sequence t, x, y  will fail (1st arg
            # is not a NumPy array)
        self._array_shape = r.shape

    # The problem with this class is that, in the vectorized version,
    # the array shape is determined in the first call, i.e., later
    # calls may return an array with the wrong shape if the shape of
    # the input arguments change! Sometimes, when called along boundaries
    # of grids, the shape may change so the next implementation is
    # slower and safer.
    
class WrapNo2Callable:
    """Turn a number (constant) into a callable function."""
    def __init__(self, constant):
        self.constant = constant

    def __call__(self, *args):
        """
        >>> w = WrapNo2Callable(4.4)
        >>> w(99)
        4.4000000000000004
        >>> # try vectorized computations:
        >>> x = linspace(1, 4, 4)
        >>> y = linspace(1, 2, 2)
        >>> xv = x[:,NewAxis]; yv = y[NewAxis,:]
        >>> xv + yv
        array([[ 2.,  3.],
               [ 3.,  4.],
               [ 4.,  5.],
               [ 5.,  6.]])
        >>> w(xv, yv)
        array([[ 4.4,  4.4],
               [ 4.4,  4.4],
               [ 4.4,  4.4],
               [ 4.4,  4.4]])

        If you want to call such a function object with space-time
        arguments and vectorized expressions, make sure the time
        argument is not the first argument. That is,
        w(xv, yv, t) is fine, but w(t, xv, yv) will return 4.4,
        not the desired array!
               
        """
        if isinstance(args[0], (float, int, complex)):
            # scalar version:
            return self.constant
        else:
            # vectorized version:
            r = args[0].copy()
            # to get right dimension of the return array,
            # compute with args in a simple formula (sum of args)
            for a in args[1:]:
                r = r + a  # in-place r+= won't work
                # (handles x,y,t - the last t just adds a constant)
            r[:] = self.constant
            return r


class WrapDiscreteData2Callable:
    """
    Turn discrete data on a uniform grid into a callable function,
    i.e., equip the data with an interpolation function.

    >>> x = linspace(0, 1, 11)
    >>> y = 1+2*x
    >>> f = WrapDiscreteData2Callable((x,y))
    >>> # or just use the wrap2callable generic function:
    >>> f = wrap2callable((x,y))
    >>> f(0.5)   # evaluate f(x) by interpolation
    1.5
    >>> f(0.5, 0.1)  # discrete data with extra time prm: f(x,t)
    1.5
    """
    def __init__(self, data):
        self.data = data  # (x,y,f) data for an f(x,y) function
        from Scientific.Functions.Interpolation \
             import InterpolatingFunction # from ScientificPython
        self.interpolating_function = \
             InterpolatingFunction(self.data[:-1], self.data[-1])
        self.ndims = len(self.data[:-1])  # no of spatial dim.
        
    def __call__(self, *args):
        # allow more arguments (typically time) after spatial pos.:
        args = args[:self.ndims]
        # args can be tuple of scalars (point) or tuple of vectors
        if isinstance(args[0], (float, int, complex)):
            return self.interpolating_function(*args)
        else:
            # args is tuple of vectors; Interpolation must work
            # with one point at a time:
            r = [self.interpolating_function(*a) for a in zip(*args)]
            return array(r)  # wrap in NumPy array

        
def wrap2callable(f, **kwargs):
    """
    Allow constants, string formulas, discrete data points,
    user-defined functions and (callable) classes to be wrapped
    in a new callable function. That is, all the mentioned data
    structures can be used as a function, usually of space and/or
    time.
    (kwargs is used for string formulas)

    >>> f1 = wrap2callable(2.0)
    >>> f1(0.5)
    2.0
    >>> f2 = wrap2callable('1+2*x')
    >>> f2(0.5)
    2.0
    >>> f3 = wrap2callable('1+2*t', independent_variable='t')
    >>> f3(0.5)
    2.0
    >>> f4 = wrap2callable('a+b*t')
    >>> f4(0.5)
    Traceback (most recent call last):
    ...
    NameError: name 'a' is not defined
    >>> f4 = wrap2callable('a+b*t', independent_variable='t', \
                           a=1, b=2)
    >>> f4(0.5)
    2.0

    >>> x = linspace(0, 1, 3); y=1+2*x
    >>> f5 = wrap2callable((x,y))
    >>> f5(0.5)
    2.0
    >>> def myfunc(x):  return 1+2*x
    >>> f6 = wrap2callable(myfunc)
    >>> f6(0.5)
    2.0
    >>> f7 = wrap2callable(lambda x: 1+2*x)
    >>> f7(0.5)
    2.0
    >>> class MyClass:
            'Representation of a function f(x; a, b) =a + b*x'
            def __init__(self, a=1, b=1):
                self.a = a;  self.b = b
            def __call__(self, x):
                return self.a + self.b*x
    >>> myclass = MyClass(a=1, b=2)
    >>> f8 = wrap2callable(myclass)
    >>> f8(0.5)
    2.0
    >>> # 3D functions:
    >>> f9 = wrap2callable('1+2*x+3*y+4*z', \
                           independent_variables=('x','y','z'))
    >>> f9(0.5,1/3.,0.25)
    4.0
    >>> # discrete 3D data:
    >>> y = linspace(0, 1, 3); z = linspace(-1, 0.5, 16)
    >>> xv = reshape(x, (len(x),1,1))
    >>> yv = reshape(y, (1,len(y),1))
    >>> zv = reshape(z, (1,1,len(z)))
    >>> def myfunc3(x,y,z):  return 1+2*x+3*y+4*z

    >>> values = myfunc3(xv, yv, zv)
    >>> f10 = wrap2callable((x, y, z, values))
    >>> f10(0.5, 1/3., 0.25)
    4.0

    One can also check what the object is wrapped as and do more
    specific operations, e.g.,
    
    >>> f9.__class__.__name__
    'StringFunction'
    >>> str(f9)     # look at function formula
    '1+2*x+3*y+4*z'
    >>> f8.__class__.__name__
    'MyClass'
    >>> f8.a, f8.b  # access MyClass-specific data
    (1, 2)

    Troubleshooting regarding string functions:
    If you use a string formula with a NumPy array, you typically get
    error messages like::
        
       TypeError: only rank-0 arrays can be converted to Python scalars.
    
    You must then make the right import (numpy is recommended)::

       from Numeric/numarray/numpy/scitools.numpytools import *
       
    in the calling code and supply the keyword argument::
        
       globals=globals()
       
    to wrap2callable. See also the documentation of class StringFunction
    for more information.
    """
    if isinstance(f, str):
        return StringFunction(f, **kwargs)
        # this is a considerable optimization (up to a factor of 3),
        # but then the additional info in the StringFunction instance
        # is lost in the calling code:
        # return StringFunction(f, **kwargs).__call__
    elif isinstance(f, (float, int, complex)):
        return WrapNo2Callable(f)
    elif isinstance(f, (list,tuple)):
        return WrapDiscreteData2Callable(f)
    elif operator.isCallable(f):
        return f
    else:
        raise TypeError, 'f of type %s is not callable' % type(f)


# problem: setitem in ArrayGen does not support multiple indices
# relying on inherited __setitem__ works fine

def NumPy_array_iterator(a, **kwargs):
    """
    Iterate over all elements in a NumPy array a.
    Return values: generator function and the code of this function.
    The numpy.ndenumerate iterator performs the same iteration over
    an array, but NumPy_array_iterator has some additional features
    (especially handy for coding finite difference stencils, see next
    paragraph).

    The keyword arguments specify offsets in the start and stop value
    of the index in each dimension. Legal values are
    offset0_start, offset0_stop, offset1_start, offset1_stop, etc.
    Also offset_start and offset_stop are legal keyword arguments,
    these imply the same offset value for all dimensions.

    Another keyword argument is no_value, which can be True or False.
    If the value is True, the iterator returns the indices as a tuple,
    otherwise (default) the iterator returns a two-tuple consisting of
    the value of the array and the corresponding indices (as a tuple).
    
    Examples::
    
    >>> q = linspace(1, 2*3*4, 2*3*4);  q.shape = (2,3,4)
    >>> it, code = NumPy_array_iterator(q)
    >>> print code  # generator function with 3 nested loops:
    def nested_loops(a):
        for i0 in xrange(0, a.shape[0]-0):
            for i1 in xrange(0, a.shape[1]-0):
                for i2 in xrange(0, a.shape[2]-0):
                    yield a[i0, i1, i2], (i0, i1, i2)
    >>> type(it)
    <type 'function'>
    >>> for value, index in it(q):
    ...     print 'a%s = %g' % (index, value)
    ...
    a(0, 0, 0) = 1
    a(0, 0, 1) = 2
    a(0, 0, 2) = 3
    a(0, 0, 3) = 4
    a(0, 1, 0) = 5
    a(0, 1, 1) = 6
    a(0, 1, 2) = 7
    a(0, 1, 3) = 8
    a(0, 2, 0) = 9
    a(0, 2, 1) = 10
    a(0, 2, 2) = 11
    a(0, 2, 3) = 12
    a(1, 0, 0) = 13
    a(1, 0, 1) = 14
    a(1, 0, 2) = 15
    a(1, 0, 3) = 16
    a(1, 1, 0) = 17
    a(1, 1, 1) = 18
    a(1, 1, 2) = 19
    a(1, 1, 3) = 20
    a(1, 2, 0) = 21
    a(1, 2, 1) = 22
    a(1, 2, 2) = 23
    a(1, 2, 3) = 24

    Here is the version where only the indices and no the values
    are returned by the iterator::

    >>> q = linspace(1, 1*3, 3);  q.shape = (1,3)
    >>> it, code = NumPy_array_iterator(q, no_value=True)
    >>> print code
    def nested_loops(a):
        for i0 in xrange(0, a.shape[0]-0):
            for i1 in xrange(0, a.shape[1]-0):
                yield i0, i1
    >>> for i,j in it(q):
    ...   print i,j
    0 0
    0 1
    0 2
    

    Now let us try some offsets::

    >>> it, code = NumPy_array_iterator(q, offset1_stop=1, offset_start=1)
    >>> print code
    def nested_loops(a):
        for i0 in xrange(1, a.shape[0]-0):
            for i1 in xrange(1, a.shape[1]-1):
                for i2 in xrange(1, a.shape[2]-0):
                    yield a[i0, i1, i2], (i0, i1, i2)
    >>> # note: the offsets appear in the xrange arguments
    >>> for value, index in it(q):
    ...     print 'a%s = %g' % (index, value)
    ...
    a(1, 1, 1) = 18
    a(1, 1, 2) = 19
    a(1, 1, 3) = 20
    """
    # build the code of the generator function in a text string
    # (since the number of nested loops needed to iterate over all
    # elements are parameterized through len(a.shape))
    dims = range(len(a.shape))
    offset_code1 = ['offset%d_start=0' % d for d in dims]
    offset_code2 = ['offset%d_stop=0'  % d for d in dims]
    for d in range(len(a.shape)):
        key1 = 'offset%d_start' % d
        key2 = 'offset%d_stop' % d
        if key1 in kwargs:
            offset_code1.append(key1 + '=' + str(kwargs[key1]))
        if key2 in kwargs:
            offset_code2.append(key2 + '=' + str(kwargs[key2]))
        
    for key in kwargs:
        if key == 'offset_start':
            offset_code1.extend(['offset%d_start=%d' % (d, kwargs[key]) \
                            for d in range(len(a.shape))])
        if key == 'offset_stop':
            offset_code2.extend(['offset%d_stop=%d' % (d, kwargs[key]) \
                            for d in range(len(a.shape))])

    no_value = kwargs.get('no_value', False)

    for line in offset_code1:
        exec line
    for line in offset_code2:
        exec line
    code = 'def nested_loops(a):\n'
    indentation = ' '*4
    indent = '' + indentation
    for dim in range(len(a.shape)):
        code += indent + \
        'for i%d in xrange(%d, a.shape[%d]-%d):\n' \
                % (dim, eval('offset%d_start' % dim),
                   dim, eval('offset%d_stop' % dim))
        indent += indentation
    index = ', '.join(['i%d' % d for d in range(len(a.shape))])
    if no_value:
        code += indent + 'yield ' + index
    else:
        code += indent + 'yield ' + 'a[%s]' % index + ', (' + index + ')'
    exec code
    return nested_loops, code

def compute_histogram(samples, nbins=50, piecewise_constant=True):
    """
    Given a NumPy array samples with random samples, this function
    returns the (x,y) arrays in a plot-ready version of the histogram.
    If piecewise_constant is True, the (x,y) arrays gives a piecewise
    constant curve when plotted, otherwise the (x,y) arrays gives a
    piecewise linear curve where the x coordinates coincide with the
    center points in each bin.
    """
    # old primitive code based on ScientificPython:
    #from Scientific.Statistics.Histogram import Histogram
    #h = Histogram(samples, nbins)
    #h.normalizeArea() # let h be a density (unit area)
    #print h.array[:,0], '\n', h.array[:,1]
    #return h.array[:,0], h.array[:,1]
    # new code based on numpy:
    import sys
    if 'numpy' in sys.modules:
        y0, xleft = histogram(samples, bins=nbins, normed=True)
        h = xleft[1] - xleft[0]  # bin width
        if piecewise_constant:
            x = zeros(2*len(xleft) + 2, type(xleft[0]))
            y = x.copy()
            for i in range(len(xleft)):
                x[2*i+1] = xleft[i]
                x[2*i+2] = xleft[i] + h
                y[2*i+1] = y0[i]
                y[2*i+2] = y0[i]
            y[0] = 0
            x[0] = xleft[0]
            y[-1] = 0
            y[-2] = y0[-1]
            x[-1] = xleft[-1] + h
            x[-2] = x[-1]
        else:
            x = zeros(len(xleft), type(xleft[0]))
            y = y0.copy()
            for i in range(len(xleft)):
                x[i] = xleft[i] + h/2.0
    return x, y

        

def factorial(n, method='reduce'):
    """
    Compute the factorial n! using long integers.
    Different implementations are available
    (see source code for the methods).

    Here is an efficiency comparison of the methods (computing 80!):
    reduce                    |     1.00
    lambda list comprehension |     1.70
    lambda functional         |     3.08
    plain recursive           |     5.83
    lambda recursive          |    21.73
    scipy                     |   131.18
    """
    if not isinstance(n, (int, long, float)):
        raise TypeError, 'factorial(n): n must be integer not %s' % type(n)
    n = long(n)

    if n == 0 or n == 1:
        return 1

    if method == 'plain iterative':
        f = 1
        for i in range(1, n+1):
            f *= i
        return f
    elif method == 'plain recursive':
        if n == 1:
            return 1
        else:
            return n*factorial(n-1, method)
    elif method == 'lambda recursive':
        fc = lambda n: n and fc(n-1)*long(n) or 1
        return fc(n)
    elif method == 'lambda functional':
        fc = lambda n: n<=0 or \
             reduce(lambda a,b: long(a)*long(b), xrange(1,n+1))
        return fc(n)
    elif method == 'lambda list comprehension':
        fc = lambda n: [j for j in [1] for i in range(2,n+1) \
                        for j in [j*i]] [-1]
        return fc(n)
    elif method == 'reduce':
        return reduce(operator.mul, xrange(2, n+1))
    elif method == 'scipy':
        try:
            import scipy.misc.common as sc
            return sc.factorial(n)
        except ImportError:
            print 'numpyutils.factorial: scipy is not available'
            print 'default method="reduce" is used instead'
            return reduce(operator.mul, xrange(2, n+1))
            # or return factorial(n)
    else:
        raise ValueError, 'factorial: method="%s" is not supported' % method



    
    



"""
Numba-accelerated backend for resolution calculations.

This backend uses Numba JIT compilation with parallel execution to achieve
~25x speedup over pure NumPy for resolution matrix calculations.
"""

import numpy as np
from numba import njit, prange

# Physical constants
CONVERT1 = 0.4246609 * np.pi / 60 / 180  # arcmin to radians
CONVERT2 = 2.072  # meV to k^2 conversion


@njit(parallel=True, fastmath=True, cache=True)
def resmat_numba_core(Q, W, ki_all, kf_all, thetam_all, thetaa_all,
                      s2theta_all, phi_all, alpha, beta, etam, etamv,
                      etaa, etaav, etas, etasv, horifoc, moncor, use_sample_mosaic):
    """
    Core resolution matrix calculation with Numba parallelization.

    This function computes resolution matrices for all Q,W points in parallel
    using Numba's prange for automatic thread distribution.

    Parameters
    ----------
    Q : ndarray
        Momentum transfer magnitudes, shape (npts,)
    W : ndarray
        Energy transfer values, shape (npts,)
    ki_all, kf_all : ndarray
        Incident and final wave vectors, shape (npts,)
    thetam_all, thetaa_all : ndarray
        Monochromator and analyzer Bragg angles, shape (npts,)
    s2theta_all, phi_all : ndarray
        Sample scattering angles, shape (npts,)
    alpha, beta : ndarray
        Horizontal and vertical collimations (radians), shape (4,)
    etam, etamv : float
        Monochromator mosaic (h, v) in radians
    etaa, etaav : float
        Analyzer mosaic (h, v) in radians
    etas, etasv : float
        Sample mosaic (h, v) in radians
    horifoc : int
        Horizontal focusing flag (-1 = off)
    moncor : int
        Monitor correction flag (1 = on)
    use_sample_mosaic : bool
        Whether to apply sample mosaic correction

    Returns
    -------
    R0 : ndarray
        Resolution prefactors, shape (npts,)
    RM : ndarray
        Resolution matrices, shape (4, 4, npts)
    """
    npts = len(Q)
    RM = np.zeros((4, 4, npts))
    R0 = np.zeros(npts)

    # Pre-compute constant diagonal elements
    G_diag = np.array([
        1.0 / alpha[0]**2, 1.0 / alpha[1]**2,
        1.0 / beta[0]**2, 1.0 / beta[1]**2,
        1.0 / alpha[2]**2, 1.0 / alpha[3]**2,
        1.0 / beta[2]**2, 1.0 / beta[3]**2
    ])
    F_diag = np.array([1.0 / etam**2, 1.0 / etamv**2, 1.0 / etaa**2, 1.0 / etaav**2])

    # Parallel loop over all points
    for i in prange(npts):
        ki = ki_all[i]
        kf = kf_all[i]
        thetam = thetam_all[i]
        thetaa = thetaa_all[i]
        s2theta = s2theta_all[i]
        phi = phi_all[i]
        q = Q[i]

        # Build G matrix (8x8 diagonal)
        G = np.diag(G_diag)

        # Build F matrix (4x4 diagonal)
        F = np.diag(F_diag)

        # Build A matrix (6x8)
        A = np.zeros((6, 8))
        tan_thetam = np.tan(thetam)
        tan_thetaa = np.tan(thetaa)
        A[0, 0] = ki / (2 * tan_thetam)
        A[0, 1] = -ki / (2 * tan_thetam)
        A[3, 4] = kf / (2 * tan_thetaa)
        A[3, 5] = -kf / (2 * tan_thetaa)
        A[1, 1] = ki
        A[2, 3] = ki
        A[4, 4] = kf
        A[5, 6] = kf

        # Build C matrix (4x8)
        C = np.zeros((4, 8))
        sin_thetam = np.sin(thetam)
        sin_thetaa = np.sin(thetaa)
        C[0, 0] = 0.5
        C[0, 1] = 0.5
        C[2, 4] = 0.5
        C[2, 5] = 0.5
        C[1, 2] = 1.0 / (2 * sin_thetam)
        C[1, 3] = -1.0 / (2 * sin_thetam)
        C[3, 6] = 1.0 / (2 * sin_thetaa)
        C[3, 7] = -1.0 / (2 * sin_thetaa)

        # Build B matrix (4x6)
        B = np.zeros((4, 6))
        cos_phi = np.cos(phi)
        sin_phi = np.sin(phi)
        cos_phi_s2 = np.cos(phi - s2theta)
        sin_phi_s2 = np.sin(phi - s2theta)
        B[0, 0] = cos_phi
        B[0, 1] = sin_phi
        B[0, 3] = -cos_phi_s2
        B[0, 4] = -sin_phi_s2
        B[1, 0] = -sin_phi
        B[1, 1] = cos_phi
        B[1, 3] = sin_phi_s2
        B[1, 4] = -cos_phi_s2
        B[2, 2] = 1.0
        B[2, 5] = -1.0
        B[3, 0] = 2 * CONVERT2 * ki
        B[3, 3] = -2 * CONVERT2 * kf

        # Cooper-Nathans calculation: M = B @ A @ inv(G + C.T @ F @ C) @ A.T @ B.T
        CtFC = C.T @ F @ C
        GpCtFC = G + CtFC
        GpCtFC_inv = np.linalg.inv(GpCtFC)
        HF = A @ GpCtFC_inv @ A.T

        # Handle horizontal focusing
        if horifoc > 0:
            HF_inv = np.linalg.inv(HF)
            HF_inv[4, 4] = (1.0 / (kf * alpha[2]))**2
            HF_inv[4, 3] = 0.0
            HF_inv[3, 4] = 0.0
            HF_inv[3, 3] = (tan_thetaa / (etaa * kf))**2
            HF = np.linalg.inv(HF_inv)

        Minv = B @ HF @ B.T
        M = np.linalg.inv(Minv)

        # Reorder to RM format (swap indices 2 and 3)
        idx = np.array([0, 1, 3, 2])
        for ii in range(4):
            for jj in range(4):
                RM[ii, jj, i] = M[idx[ii], idx[jj]]

        # Prefactor calculation
        Rm = ki**3 / tan_thetam
        Ra = kf**3 / tan_thetaa
        det_F = np.linalg.det(F)
        det_GpCtFC = np.linalg.det(GpCtFC)

        R0_val = Rm * Ra * (2 * np.pi)**4 / (64 * np.pi**2 * sin_thetam * sin_thetaa) \
                 * np.sqrt(det_F / det_GpCtFC)

        # Monitor correction
        if moncor == 1:
            g = G[0:4, 0:4]
            f = F[0:2, 0:2]
            c = C[0:2, 0:4]
            ctfc = c.T @ f @ c
            det_f = np.linalg.det(f)
            det_gctfc = np.linalg.det(g + ctfc)
            Rmon = Rm * (2 * np.pi)**2 / (8 * np.pi * sin_thetam) * np.sqrt(det_f / det_gctfc)
            R0_val = R0_val / Rmon * ki

        # Chesser-Axe normalization
        det_RM = np.linalg.det(RM[:, :, i])
        R0_val = R0_val / (2 * np.pi)**2 * np.sqrt(det_RM)

        # kf/ki factor
        R0_val = R0_val * kf / ki

        # Sample mosaic correction
        if use_sample_mosaic:
            R0_val = R0_val / np.sqrt(
                (1 + (q * etas)**2 * RM[3, 3, i]) *
                (1 + (q * etasv)**2 * RM[1, 1, i])
            )
            # Modify RM for sample mosaic
            RM_pt = RM[:, :, i].copy()
            RM_inv = np.linalg.inv(RM_pt)
            RM_inv[1, 1] = RM_inv[1, 1] + q**2 * etas**2
            RM_inv[3, 3] = RM_inv[3, 3] + q**2 * etasv**2
            RM_new = np.linalg.inv(RM_inv)
            for ii in range(4):
                for jj in range(4):
                    RM[ii, jj, i] = RM_new[ii, jj]

        R0[i] = R0_val

    return R0, RM


class NumbaBackend:
    """
    Numba-accelerated backend for resolution calculations.

    Provides ~25x speedup over pure NumPy by using JIT compilation
    and parallel execution across CPU cores.
    """

    name = 'numba'

    def __init__(self):
        self._compiled = False

    def _warmup(self):
        """Trigger JIT compilation with small test arrays."""
        if not self._compiled:
            # Small warmup to compile the function
            n = 10
            Q = np.linspace(0.5, 2.0, n)
            W = np.linspace(1.0, 5.0, n)
            ki = np.sqrt((14.7 + W) / CONVERT2)
            kf = np.full(n, np.sqrt(14.7 / CONVERT2))
            thetam = np.arcsin(1.87325 / (2 * ki))
            thetaa = np.arcsin(1.87325 / (2 * kf))
            cos_s2theta = np.clip((ki**2 + kf**2 - Q**2) / (2 * ki * kf), -1, 1)
            s2theta = -np.arccos(cos_s2theta)
            phi = np.arctan2(-kf * np.sin(s2theta), ki - kf * np.cos(s2theta))
            alpha = np.array([40, 40, 40, 40]) * CONVERT1
            beta = np.array([120, 120, 120, 120]) * CONVERT1
            eta = 30 * CONVERT1

            _ = resmat_numba_core(Q, W, ki, kf, thetam, thetaa, s2theta, phi,
                                  alpha, beta, eta, eta, eta, eta, eta, eta,
                                  -1, 1, True)
            self._compiled = True

    def compute_resolution(self, Q, W, EXP):
        """
        Compute resolution matrices using Numba-accelerated code.

        Parameters
        ----------
        Q : ndarray
            Momentum transfer magnitudes
        W : ndarray
            Energy transfers
        EXP : list of dict
            Experiment configuration (uses first entry)

        Returns
        -------
        R0 : ndarray
            Prefactors
        RM : ndarray
            Resolution matrices, shape (4, 4, npts)
        """
        self._warmup()

        exp0 = EXP[0]
        npts = len(Q)

        # Extract parameters
        mono = exp0['mono']
        ana = exp0['ana']
        sample = exp0.get('sample', {})

        alpha = np.asarray(exp0['hcol'], dtype=np.float64) * CONVERT1
        beta = np.asarray(exp0['vcol'], dtype=np.float64) * CONVERT1

        etam = float(mono['mosaic']) * CONVERT1
        etamv = float(mono.get('vmosaic', mono['mosaic'])) * CONVERT1
        etaa = float(ana['mosaic']) * CONVERT1
        etaav = float(ana.get('vmosaic', ana['mosaic'])) * CONVERT1

        etas = float(sample.get('mosaic', 0)) * CONVERT1 if sample else 0.0
        etasv = float(sample.get('vmosaic', sample.get('mosaic', 0))) * CONVERT1 if sample else 0.0
        use_sample_mosaic = 'mosaic' in sample if sample else False

        efixed = float(exp0['efixed'])
        infin = exp0.get('infin', -1)
        epm = exp0.get('dir1', 1)
        ep = exp0.get('dir2', 1)

        tau_list = {
            'pg(002)': 1.87325, 'pg(004)': 3.74650,
            'ge(111)': 1.92366, 'ge(220)': 3.14131,
            'ge(311)': 3.68351, 'be(002)': 3.50702,
            'pg(110)': 5.49806
        }
        taum = tau_list[mono['tau']]
        taua = tau_list[ana['tau']]

        horifoc = exp0.get('horifoc', -1)
        moncor = exp0.get('moncor', 1)

        # Compute energies and wave vectors
        Q = np.asarray(Q, dtype=np.float64)
        W = np.asarray(W, dtype=np.float64)

        if infin > 0:
            ei = np.full(npts, efixed)
            ef = efixed - W
        else:
            ei = efixed + W
            ef = np.full(npts, efixed)

        ki = np.sqrt(ei / CONVERT2)
        kf = np.sqrt(ef / CONVERT2)

        sign_epm = 1 if epm > 0 else -1
        sign_ep = 1 if ep > 0 else -1
        thetam = np.arcsin(taum / (2 * ki)) * sign_epm
        thetaa = np.arcsin(taua / (2 * kf)) * sign_ep

        cos_s2theta = (ki**2 + kf**2 - Q**2) / (2 * ki * kf)
        cos_s2theta = np.clip(cos_s2theta, -1.0, 1.0)
        s2theta = -np.arccos(cos_s2theta)
        phi = np.arctan2(-kf * np.sin(s2theta), ki - kf * np.cos(s2theta))

        # Call the Numba-compiled function
        R0, RM = resmat_numba_core(
            Q, W, ki, kf, thetam, thetaa, s2theta, phi,
            alpha, beta, etam, etamv, etaa, etaav, etas, etasv,
            horifoc, moncor, use_sample_mosaic
        )

        return R0, RM



from numpy import *

def PrefDemo(H,K,L,W,myrescal,p):
    """This is an example of a  "slowly varying prefactor and background" function for use with ConvRes.m.
    In the particular case it is a form factor squared for Ni2+.
    Two modes are assumed, polarized along the b and c axis of the crystal.
    The function also calculates the two magnetic polarization factors.
    It is meant to be used in combination with the cross section function SqwDemo.m
    All arguments are vectors, so dont forget to use ".*" instead of "*", etc."""

    I=p[4] #							% Intensity prefactor.
    bgr=p[5] #						% A flat background.

    # Here we will need to access to direct
    # and reciprocal space lattice parameters. The function GetLattice
    # can be used to get those directly from EXP.
    #[sample,rsample]=GetLattice(EXP);


    #% First, calculate the scattering vector modulus squared
    #q2=myrescal.lattice_calculator.modvec(H,K,L,'latticestar')**2

    #% Now, use the Jane Brown approximate expression for Ni2+
    #d=q2/(16*pi**2)
    #ff=0.0163*exp(-35.883*sd)+0.3916*exp(-13.223*sd)+0.6052*exp(-4.339*sd)-0.0133

    # Calculate the polarization factors for transverse excitations
    # For this we utilize the function angle2.m
    # Angle between (0,1,0) [direct space] and [h k l]:
    #alphay=myrescal.lattice_calculator.angle2(0,1,0,H,K,L)
    # Angle between (0,0,1) [direct space] and [h k l]:
    #alphaz=myrescal.lattice_calculator.angle2(0,0,1,H,K,L)
    # Angle between (1,0,0) [direct space] and [h k l]:
    #alphax=myrescal.lattice_calculator.angle2(1,0,0,H,K,L)

    # Polarization factors for each of the three modes.
    #polx=sin(alphax)**2
    #poly=sin(alphay)**2
    #polz=sin(alphaz)**2
    prefactor=ones((1,H.shape[0]),'float64')*p[4]
    #prefactor[0,:]=ff**2*polx*p[5] # %Apply the overall scaling factor here
    #prefactor[1,:]=ff**2*poly*p[5]
    #prefactor[2,:]=ff**2*polz*p[5]

    #The background will be simply a constant
    bgr=ones(H.shape,'float64')*p[5]

    return prefactor,bgr


from __future__ import division
import sys
import numpy as N
import matplotlib
matplotlib.use('Agg')
import pylab
import math
import unittest
from matplotlib.patches import Ellipse
from matplotlib.ticker import NullFormatter, MultipleLocator
from matplotlib.ticker import FormatStrFormatter
from matplotlib.ticker import MaxNLocator
import copy,sys

import lattice_calculator

eps=1e-3
pi=N.pi

def sign(x):
     if x>0:
          ret=1
     if x<0:
          ret=-1
     if x==0:
          ret=0
     return ret

def blkdiag(g):
     "Returns a block diagonal matrix, given a list of square matrices, g"
     glen=len(g)
     n=0
     for i in range(glen):
     #     print g[i]
          n=n+g[i].shape[0]
     #     print n


     gout=N.zeros((n,n))
     offset=0
     for i in range(glen):
          currblock=g[i]
          lenx,leny=currblock.shape
     #    print i
     #    print lenx, leny
          for x in range(lenx):
               for y in range(leny):
                    gout[x+offset,y+offset]=currblock[x,y]
     #           print gout
          offset=offset+lenx
     return gout


def similarity_transform(A,B):
     G=N.dot(B,A.transpose())
     G2=N.dot(A,G)
     return G2

class rescalculator:
     def __init__(self, mylattice):
          self.lattice_calculator=mylattice
          return

     def ResMat(self, Q, W,EXP):
          CONVERT1=0.4246609*N.pi/60/180;
          CONVERT2=2.072;
          npts=len(EXP)
          Qnpts=N.size(Q)
          RM=N.zeros((4, 4, npts),'d');
          R0=N.zeros((npts,1),'d');
          RM_=N.zeros((4, 4),'d');
          D=N.zeros((8, 13),'d');
          d=N.zeros((4, 7),'d');
          T=N.zeros((4, 13),'d');
          t=N.zeros((2, 7),'d');
          A=N.zeros((6, 8),'d');
          C=N.zeros((4, 8),'d');
          B=N.zeros((4, 6),'d');

          for ind in range(npts):
#            %Assign default values and decode parameters
               moncor=1;
               if 'moncor' in EXP[ind]:
                    moncor = EXP[ind]['moncor']
               alpha = EXP[ind]['hcol']*CONVERT1
               beta =  EXP[ind]['vcol']*CONVERT1
               mono=EXP[ind]['mono']
               etam = mono['mosaic']*CONVERT1
               etamv=etam
               if 'vmosaic' in mono:
                    etamv = mono['vmosaic']*CONVERT1
               ana=EXP[ind]['ana']
               etaa = ana['mosaic']*CONVERT1
               etaav=etaa
               if 'vmosaic' in ana:
                    etaav = ana['vmosaic']*CONVERT1
               sample=EXP[ind]['sample']
               infin=-1
               if 'infin' in EXP[ind]:
                    infin = EXP[ind]['infin']
               efixed=EXP[ind]['efixed']
               epm=1
               if 'dir1' in EXP[ind]:
                    epm= EXP[ind]['dir1']
               ep=1
               if 'dir2' in EXP[ind]:
                    ep= EXP[ind]['dir2']
               monitorw=1.0
               monitorh=1.0
               beamw=1.0
               beamh=1.0
               monow=1.0
               monoh=1.0
               monod=1.0
               anaw=1.0
               anah=1.0
               anad=1.0
               detectorw=1.0
               detectorh=1.0
               sshape=N.eye(3)
               L0=1.0
               L1=1.0
               L1mon=1.0
               L2=1.0
               L3=1.0
               monorv=1e6
               monorh=1e6
               anarv=1e6
               anarh=1e6
               if 'beam' in EXP[ind]:
                    beam=EXP[ind]['beam']
                    if 'width' in beam:
                         beamw=beam['width']**2
                    if 'height' in beam:
                         beamh=beam['height']**2
               bshape=N.diag([beamw,beamh])
               if 'monitor' in EXP[ind]:
                    monitor=EXP[ind]['monitor']
                    if 'width' in monitor:
                         monitorw=monitor['width']**2
                    monitorh=monitorw
                    if 'height' in monitor:
                         monitorh=monitor['height']**2;
               monitorshape=N.diag([monitorw,monitorh])
               if 'detector' in EXP[ind]:
                    detector=EXP[ind]['detector']
                    if 'width' in detector:
                         detectorw=detector['width']**2
                    if 'height' in detector:
                         detectorh=detector['height']**2
               dshape=N.diag([detectorw,detectorh])
               if 'width' in mono:
                    monow=mono['width']**2
               if 'height' in mono:
                    monoh=mono['height']**2
               if 'depth' in mono:
                    monod=mono.depth**2
               mshape=N.diag([monod,monow,monoh])
               if 'width' in ana:
                    anaw=ana['width']**2
               if 'height' in ana:
                    anah=ana['height']**2
               if 'depth' in ana:
                    anad=ana['depth']**2
               ashape=N.diag([anad,anaw,anah])
               if 'shape' in sample:
                    sshape=sample['shape']
               if 'arms' in EXP[ind]:
                    arms=EXP[ind]['arms']
                    L0=arms[0]
                    L1=arms[2]
                    L2=arms[3]
                    L3=arms[4]
                    L1mon=L1
                    if len(arms)>3:
                         L1mon=arms[4]
               if 'rv' in mono:
                    monorv=mono['rv']
               if 'rh' in mono:
                    monorh=mono['rh']
               if 'rv' in ana:
                    anarv=ana['rv']
               if 'rh' in ana:
                    anarh=ana['rh']
               method=0;
               if 'method' in EXP[ind]:
                    method=EXP[ind]['method']
               myinstrument=lattice_calculator.Instrument()
               taum=myinstrument.get_tau(mono['tau'])
               taua=myinstrument.get_tau(ana['tau'])

               horifoc=-1;
               if 'horifoc' in EXP[ind]:
                    horifoc=EXP[ind]['horifoc']
               if horifoc==1:
                    alpha[2]=alpha[2]*N.sqrt(8*N.log(2)/12)
#            %---------------------------------------------------------------------------------------------
#            %Calculate angles and energies
               w=W[ind]
               q=Q[ind]
               ei=efixed
               ef=efixed
               if infin>0:
                    ef=efixed-w
               else:
                    ei=efixed+w
               ki = N.sqrt(ei/CONVERT2)
               kf = N.sqrt(ef/CONVERT2)
               thetam=N.arcsin(taum/(2*ki))*sign(epm)
               thetaa=N.arcsin(taua/(2*kf))*sign(ep)
               s2theta=-N.arccos( (ki**2+kf**2-q**2)/(2*ki*kf))# %2theta sample
               thetas=s2theta/2
               phi=N.arctan2(-kf*N.sin(s2theta), ki-kf*N.cos(s2theta)) #%Angle from ki to Q

     #           %---------------------------------------------------------------------------------------------
#            %Calculate beam divergences defined by neutron guides
               pi=N.pi
               if alpha[0]<0:
                    alpha[0]=-alpha[0]*2*0.427/ki*pi/180
               if alpha[1]<0:
                    alpha[1]=-alpha[1]*2*0.427/ki*pi/180
               if alpha[2]<0:
                    alpha[2]=-alpha[2]*2*0.427/ki*pi/180
               if alpha[3]<0:
                    alpha[3]=-alpha[3]*2*0.427/ki*pi/180

               if beta[0]<0:
                    beta[0]=-beta[0]*2*0.427/ki*pi/180
               if beta[1]<0:
                    beta[1]=-beta[1]*2*0.427/ki*pi/180
               if beta[2]<0:
                    beta[2]=-beta[2]*2*0.427/ki*pi/180
               if beta[3]<0:
                    beta[3]=-beta[3]*2*0.427/ki*pi/180

#            %---------------------------------------------------------------------------------------------
#            %Rededine sample geometry
               psi=thetas-phi;# %Angle from sample geometry X axis to Q
               rot=N.zeros((3,3))
               rot[0,0]=N.cos(psi)
               rot[1,1]=N.cos(psi)
               rot[0,1]=N.sin(psi)
               rot[1,0]=-N.sin(psi)
               rot[2,2]=1
               sshape=N.dot(rot.transpose(),N.dot(sshape,rot)) #matrix multiplication?
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix G
               G=1.0/N.array([alpha[0],alpha[1],beta[0],beta[1],alpha[2],alpha[3],beta[2],beta[3]])**2
               G=N.diag(G)
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix F
               F=1.0/N.array([etam,etamv,etaa,etaav])**2
               F=N.diag(F);
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix A
               A[0,0]=ki/2/N.tan(thetam)
               A[0,1]=-A[0,0]
               A[3,4]=kf/2/N.tan(thetaa)
               A[3,5]=-A[3,4]
               A[1,1]=ki
               A[2,3]=ki
               A[4,4]=kf
               A[5,6]=kf
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix C
               C[0,0]=1.0/2
               C[0,1]=1.0/2
               C[2,4]=1.0/2
               C[2,5]=1.0/2
               C[1,2]=1.0/(2*N.sin(thetam))
               C[1,3]=-C[1,2];# %mistake in paper
               C[3,6]=1.0/(2*N.sin(thetaa))
               C[3,7]=-C[3,6]
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix B
               B[0,0]=N.cos(phi)
               B[0,1]=N.sin(phi)
               B[0,3]=-N.cos(phi-s2theta)
               B[0,4]=-N.sin(phi-s2theta)
               B[1,0]=-B[0,1]
               B[1,1]=B[0,0]
               B[1,3]=-B[0,4]
               B[1,4]=B[0,3]
               B[2,2]=1.0
               B[2,5]=-1.0
               B[3,0]=2*CONVERT2*ki
               B[3,3]=-2*CONVERT2*kf
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix S
               Sinv=blkdiag([bshape,mshape,sshape,ashape,dshape])# %S-1 matrix
               S=N.linalg.inv(Sinv)
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix T
               T[0,0]=-1./(2*L0); # %mistake in paper
               T[0,2]=N.cos(thetam)*(1./L1-1./L0)/2
               T[0,3]=N.sin(thetam)*(1./L0+1./L1-2./(monorh*N.sin(thetam)))/2
               T[0,5]=N.sin(thetas)/(2*L1)
               T[0,6]=N.cos(thetas)/(2*L1)
               T[1,1]=-1./(2*L0*N.sin(thetam))
               T[1,4]=(1./L0+1./L1-2*N.sin(thetam)/monorv)/(2*N.sin(thetam))
               T[1,7]=-1./(2*L1*N.sin(thetam))
               T[2,5]=N.sin(thetas)/(2*L2)
               T[2,6]=-N.cos(thetas)/(2*L2)
               T[2,8]=N.cos(thetaa)*(1./L3-1./L2)/2
               T[2,9]=N.sin(thetaa)*(1./L2+1./L3-2/(anarh*N.sin(thetaa)))/2
               T[2,11]=1./(2*L3)
               T[3,7]=-1./(2*L2*N.sin(thetaa))
               T[3,10]=(1./L2+1./L3-2*N.sin(thetaa)/anarv)/(2*N.sin(thetaa))
               T[3,12]=-1./(2*L3*N.sin(thetaa))
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix D
#            % Lots of index mistakes in paper for matix D
               D[0,0]=-1./L0
               D[0,2]=-N.cos(thetam)/L0
               D[0,3]=N.sin(thetam)/L0
               D[2,1]=D[0,0]
               D[2,4]=-D[0,0]
               D[1,2]=N.cos(thetam)/L1
               D[1,3]=N.sin(thetam)/L1
               D[1,5]=N.sin(thetas)/L1
               D[1,6]=N.cos(thetas)/L1
               D[3,4]=-1./L1
               D[3,7]=-D[3,4]
               D[4,5]=N.sin(thetas)/L2
               D[4,6]=-N.cos(thetas)/L2
               D[4,8]=-N.cos(thetaa)/L2
               D[4,9]=N.sin(thetaa)/L2
               D[6,7]=-1./L2
               D[6,10]=-D[6,7]
               D[5,8]=N.cos(thetaa)/L3
               D[5,9]=N.sin(thetaa)/L3
               D[5,11]=1./L3
               D[7,10]=-D[5,11]
               D[7,12]=D[5,11]
#            %---------------------------------------------------------------------------------------------
#            %Definition of resolution matrix M
               if method==1:
                    Minv=N.dot(N.dot(B,A),N.dot(N.linalg.inv(D.linalg.inv(D*N.linalg.inv(S+T.transpose()*F*T)*D.transpose())+G),\
                                                N.dot(A.transpose(),B.transpose())))#; %Popovici
                    #Minv=B*A*((D*(S+T'*F*T)^(-1)*D')^(-1)+G)^(-1)*A'*B'; %Popovici
               else:
                    #%Horizontally focusing analyzer if needed
                    #print 'intermediate C.T*F*C'
                    #print N.dot(C.transpose(),N.dot(F,C))
                    #print 'product'
                    #print N.dot(F,C)
                    HF_int=N.linalg.inv(G+N.dot(C.transpose(),N.dot(F,C)))
                    HF=similarity_transform(A,HF_int)
                    #print 'HF'
                    #print HF
                    if horifoc>0:
                         HF=N.linalg.inv(HF);
                         HF[4,4]=(1.0/(kf*alpha[2]))**2
                         HF[4,3]=0
                         HF[3,4]=0
                         HF[3,3]=(N.tan(thetaa)/(etaa*kf))**2
                         HF=N.linalg.inv(HF)
                    Minv=similarity_transform(B,HF)#; %Cooper-Nathans
               M=N.linalg.inv(Minv)
               #print 'A'
               #print A
               #print 'B'
               #print B
               #print 'C'
               #print C
               #print 'D'
               #print D
               #print 'T'
               #print T
               #print 'M'
               #print M
               #print 'Minv'
               #print Minv
               #print 'G'
               #print G
               #print 'F'
               #print F
               RM_[0,0]=M[0,0]
               RM_[1,0]=M[1,0]
               RM_[0,1]=M[0,1]
               RM_[1,1]=M[1,1]

               RM_[0,2]=M[0,3]
               RM_[2,0]=M[3,0]
               RM_[2,2]=M[3,3]
               RM_[2,1]=M[3,1]
               RM_[1,2]=M[1,3]

               RM_[0,3]=M[0,2]
               RM_[3,0]=M[2,0]
               RM_[3,3]=M[2,2]
               RM_[3,1]=M[2,1]
               RM_[1,3]=M[1,2]
     #           %--------------------------------------------------------------------------------------------
     #           %Calculation of prefactor, normalized to source
               #print 'RM_'
               #print RM_
               Rm=ki**3/N.tan(thetam)
               Ra=kf**3/N.tan(thetaa)
               #print 'Rm'
               #print Rm
               #print 'Ra'
               #print Ra
               if method==1:
                    R0_=Rm*Ra*(2*pi)**4/(64*pi**2*N.sin(thetam)*N.sin(thetaa))\
                       *N.sqrt(N.linalg.det(F)/N.linalg.det\
                               (N.linalg.inv(N.dot(D,N.dot(N.linalg.inv(S+N.dot(T.T,N.dot(F,T))),D.T)))+G)) #%Popovici
               else:
                    R0_=Rm*Ra*(2*pi)**4/(64*pi**2*N.sin(thetam)*N.sin(thetaa))\
                       *N.sqrt( N.linalg.det(F)/N.linalg.det(G+N.dot(C.transpose(),N.dot(F,C)))) #%Cooper-Nathans
                    #print 'RO_'
                    #print R0_
#            %---------------------------------------------------------------------------------------------
#            %Normalization to flux on monitor
               if moncor==1:
                    #print 'Moncor'
                    #print R0_
                    #print 'RM in Moncor'
                    #print RM_
                    g=G[0:4][:,0:4]
                    f=F[0:2][:,0:2]
                    c=C[0:2][:,0:4]
                    #print 'f'
                    #print f
                    #print 'g'
                    #print g
                    #print 'c'
                    #print c

                    t[0,0]=-1./(2*L0)#  %mistake in paper
                    t[0,2]=N.cos(thetam)*(1./L1mon-1./L0)/2
                    t[0,3]=N.sin(thetam)*(1./L0+1./L1mon-2./(monorh*N.sin(thetam)))/2
                    t[0,6]=1./(2*L1mon)
                    t[1,1]=-1./(2*L0*N.sin(thetam))
                    t[1,4]=(1./L0+1./L1mon-2*N.sin(thetam)/monorv)/(2*N.sin(thetam))
                    sinv=blkdiag([bshape,mshape,monitorshape])# %S-1 matrix
                    s=N.linalg.inv(sinv)
                    d[0,0]=-1./L0
                    d[0,2]=-N.cos(thetam)/L0
                    d[0,3]=N.sin(thetam)/L0
                    d[2,1]=D[0,0]
                    d[2,4]=-D[0,0]
                    d[1,2]=N.cos(thetam)/L1mon
                    d[1,3]=N.sin(thetam)/L1mon
                    d[1,5]=0
                    d[1,6]=1./L1mon
                    d[3,4]=-1./L1mon
                    if method==1:
                         Rmon=Rm*(2*pi)**2/(8*pi*N.sin(thetam))*N.sqrt(N.linalg.det(f)/\
                                                                       N.linalg.det(N.linalg.inv(N.dot(d,N.dot(N.linalg.inv(s+t.transpose()*f*t),d.transpose())))+g)) #%Popovici
                    else:
                         Rmon=Rm*(2*pi)**2/(8*pi*N.sin(thetam))*N.sqrt(N.linalg.det(f)/N.linalg.det(g+N.dot(c.transpose(),N.dot(f,c)))) #%Cooper-Nathans
                         #print 'Cooper Nathans'
                         #print 'Rmon', Rmon
                    R0_=R0_/Rmon
                    R0_=R0_*ki #%1/ki monitor efficiency
                    #print 'R01', R0_
                    #print 'Rmon', Rmon

#            %---------------------------------------------------------------------------------------------
#            %Transform prefactor to Chesser-Axe normalization
               R0_=R0_/(2*pi)**2*N.sqrt(N.linalg.det(RM_))
#            %---------------------------------------------------------------------------------------------
#            %Include kf/ki part of cross section
               R0_=R0_*kf/ki
#            %---------------------------------------------------------------------------------------------
#            %Take care of sample mosaic if needed [S. A. Werner & R. Pynn, J. Appl. Phys. 42, 4736, (1971)]
               #print 'R0_ before mosaic', R0_
               #print RM_
               if 'mosaic' in sample:
                    etas = sample['mosaic']*CONVERT1;
                    #print 'sample',sample
                    if 'vmosaic' in sample:
                         #print 'vertical'
                         etasv = sample['vmosaic']*CONVERT1
                         #print 'etasv',etasv
                    else:
                         etasv=etas

                    #print 'etas',etas, 'etasv',etasv
                    R0_=R0_/N.sqrt((1.+(q*etas)**2*RM_[3,3])*(1.0+(q*etasv)**2*RM_[1,1]))
                    Minv=N.linalg.inv(RM_)
                    Minv[1,1]=Minv[1,1]+q**2*etas**2
                    Minv[3,3]=Minv[3,3]+q**2*etasv**2
                    RM_=N.linalg.inv(Minv)
                    #print 'mosaic'
                    #print 'R0_', R0_
                    #print 'RM_'
                    #print RM_
#            %---------------------------------------------------------------------------------------------
#            %Take care of analyzer reflectivity if needed [I. Zaliznyak, BNL]
               if ('thickness' in ana) & ('Q' in ana):
                    KQ = ana['Q']
                    KT = ana['thickness']
                    toa=(taua/2)/N.sqrt(kf**2-(taua/2)**2)
                    smallest=alpha[3]
                    if alpha[3]>alpha[2]:
                         smallest=alpha[2]
                    Qdsint=KQ*toa
                    dth=(N.arange(201)/200)*N.sqrt(2*N.log(2))*smallest
                    wdth=N.exp(-dth**2/2./etaa**2)
                    sdth=KT*Qdsint*wdth/etaa/N.sqrt(2.*pi)
                    rdth=1./(1+1./sdth)
                    reflec=rdth.sum()/wdth.sum()
                    R0_=R0_*reflec
#            %---------------------------------------------------------------------------------------------
               #print 'ind ', ind
               #print 'shape ', R0.shape
               R0[ind]=R0_
               RM[:,:,ind]=RM_[:,:]
          return R0, RM

     def ResMatS(self,H,K,L,W,EXP):
# [len,H,K,L,W,EXP]=CleanArgs(H,K,L,W,EXP);
          x=self.lattice_calculator.x
          y=self.lattice_calculator.y
          z=self.lattice_calculator.z
          Q=self.lattice_calculator.modvec(H,K,L,'latticestar')
          uq=N.zeros((3,self.lattice_calculator.npts),'d')
          uq[0,:]=H/Q;  #% Unit vector along Q
          uq[1,:]=K/Q;
          uq[2,:]=L/Q;
          xq=self.lattice_calculator.scalar(x[0,:],x[1,:],x[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
          yq=self.lattice_calculator.scalar(y[0,:],y[1,:],y[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
          zq=0; # %scattering vector assumed to be in (self.orient1,self.orient2) plane;
          tmat=N.zeros((4,4,self.lattice_calculator.npts)); #%Coordinate transformation matrix
          tmat[3,3,:]=1;
          tmat[2,2,:]=1;
          tmat[0,0,:]=xq;
          tmat[0,1,:]=yq;
          tmat[1,1,:]=xq;
          tmat[1,0,:]=-yq;

          RMS=N.zeros((4,4,self.lattice_calculator.npts));
          rot=N.zeros((3,3));
          EXProt=EXP;

#        %Sample shape matrix in coordinate system defined by scattering vector
          for i in range(len(EXP)):
               sample=EXP[i]['sample'];
               if 'shape' in sample:
                    rot[0,0]=tmat[0,0,i];
                    rot[1,0]=tmat[1,0,i];
                    rot[0,1]=tmat[0,1,i];
                    rot[1,1]=tmat[1,1,i];
                    rot[2,2]=tmat[2,2,i];
                    EXProt[i]['sample']['shape']=N.dot(rot,N.dot(sample['shape'],rot.T));

          R0,RM= self.ResMat(Q,W,EXProt)
          #print 'RM ',RM.shape
          #print 'npts ',self.lattice_calculator.npts
          #print 'tmat ',tmat.shape
          for i in range(self.lattice_calculator.npts):
               RMS[:,:,i]=N.dot((tmat[:,:,i]).transpose(),N.dot(RM[:,:,i],tmat[:,:,i]))

          mul=N.zeros((4,4));
          e=N.eye(4,4);
          for i in range(len(EXP)):
               if 'Smooth' in EXP[i]:
                    if 'X' in (EXP[i]['Smooth']):
                         mul[0,0]=1./(EXP[i]['Smooth']['X']**2/8/N.log(2));
                         mul[1,1]=1./(EXP[i]['Smooth']['Y']**2/8/N.log(2));
                         mul[2,2]=1./(EXP[i]['Smooth']['E']**2/8/N.log(2));
                         mul[3,3]=1./(EXP[i]['Smooth']['Z']**2/8/N.log(2));
                         R0[i]=R0[i]/N.sqrt(N.linalg.det(e/RMS[:,:,i]))*N.sqrt(N.linalg.det(e/mul+e/RMS[:,:,i]));
                         RMS[:,:,i]=e/(e/mul+e/RMS[:,:,i]);
          return R0.T, RMS

     def ResPlot(self,H,K,L,W,EXP):
          """Plot resolution ellipse for a given scan"""
          center=N.round(H.shape[0]/2)
          if center<1:
               center=0
          if center>H.shape[0]:
               center=H.shape[0]
          #EXP=[EXP[center]]
          Style1=''
          Style2='--'

          XYAxesPosition=[0.1, 0.6, 0.3, 0.3]
          XEAxesPosition=[0.1, 0.1, 0.3, 0.3]
          YEAxesPosition=[0.6, 0.6, 0.3, 0.3]
          TextAxesPosition=[0.45, 0.0, 0.5, 0.5]
          GridPoints=101

          [R0,RMS]=self.ResMatS(H,K,L,W,EXP)
          #[xvec,yvec,zvec,sample,rsample]=self.StandardSystem(EXP);
          self.lattice_calculator.StandardSystem()
          #print 'shape ',self.lattice_calculator.x.shape
          qx=self.lattice_calculator.scalar(self.lattice_calculator.x[0,:],self.lattice_calculator.x[1,:],self.lattice_calculator.x[2,:],H,K,L,'latticestar')
          qy=self.lattice_calculator.scalar(self.lattice_calculator.y[0,:],self.lattice_calculator.y[1,:],self.lattice_calculator.y[2,:],H,K,L,'latticestar')
          qw=W;

          #========================================================================================================
          #find reciprocal-space directions of X and Y axes

          o1=self.lattice_calculator.orientation.orient1.T#[:,0] #EXP['orient1']
          o2=self.lattice_calculator.orientation.orient2.T#[:,0] #EXP['orient2']
          pr=self.lattice_calculator.scalar(o2[0,:],o2[1,:],o2[2,:],self.lattice_calculator.y[0,:],self.lattice_calculator.y[1,:],self.lattice_calculator.y[2,:],'latticestar')
          o2[0]=self.lattice_calculator.y[0,:]*pr
          o2[1]=self.lattice_calculator.y[1,:]*pr
          o2[2]=self.lattice_calculator.y[2,:]*pr

          if N.abs(o2[0,center])<1e-5:
               o2[0,center]=0.0
          if N.absolute(o2[1,center])<1e-5:
               o2[1,center]=0.0
          if N.absolute(o2[2,center])<1e-5:
               o2[2,center]=0.0

          if N.abs(o1[0,center])<1e-5:
               o1[0,center]=0.0
          if N.absolute(o1[1,center])<1e-5:
               o1[1,center]=0.0
          if N.absolute(o1[2,center])<1e-5:
               o1[2,center]=0.0

          #%========================================================================================================
          #%determine the plot range
          XWidth=max(self.fproject(RMS,0))
          YWidth=max(self.fproject(RMS,1))
          WWidth=max(self.fproject(RMS,2))
          XMax=(max(qx)+XWidth*1.5)
          XMin=(min(qx)-XWidth*1.5)
          YMax=(max(qy)+YWidth*1.5)
          YMin=(min(qy)-YWidth*1.5)
          WMax=(max(qw)+WWidth*1.5)
          WMin=(min(qw)-WWidth*1.5)
          #print 'qx ',qx
          #print 'qy ',qy
          #print 'XWidth ',XWidth
          #print 'YWidth ',YWidth
          fig=pylab.figure()
          #%========================================================================================================
          #% plot XY projection


          proj,sec=self.project(RMS,2)
          (a,b,c)=N.shape(proj)
          mat=N.copy(proj)
          #print 'proj ', proj.shape
          a1=[];b1=[];theta=[];a1_sec=[];b1_sec=[];theta_sec=[];e=[]; e_sec=[]
          for i in range(c):
               matm=N.matrix(mat[:,:,i])
               w,v=N.linalg.eig(matm)
               vm=N.matrix(v)
               vmt=vm.T
               mat_diag=vmt*matm*vm
               a1.append(1.0/N.sqrt(mat_diag[0,0]))
               b1.append(1.0/N.sqrt(mat_diag[1,1]))
               #thetar=N.arccos(vm[0,0])#*N.sign(vm[0,0])
               thetar=-N.arctan2(vm[0,1],vm[0,0])
               theta.append(math.degrees(thetar))

          mat_sec=N.copy(sec)
          #print 'proj ', proj
          (a,b,c)=N.shape(sec)
          for i in range(c):
               matm_sec=N.matrix(mat_sec[:,:,i])
               w_sec,v_sec=N.linalg.eig(matm_sec)
               vm_sec=N.matrix(v)
               vmt_sec=vm_sec.T
               mat_diag_sec=vmt_sec*matm_sec*vm_sec
               a1_sec.append(1.0/N.sqrt(mat_diag_sec[0,0]))
               b1_sec.append(1.0/N.sqrt(mat_diag_sec[1,1]))
               #thetar_sec=N.arccos(vm_sec[0,0])*N.sign(vm_sec[0,0])
               thetar_sec=-N.arctan2(vm_sec[0,1],vm_sec[0,0])
               theta_sec.append(math.degrees(thetar_sec))
               x0y0=N.array([qx[i],qy[i]])
               e.append(Ellipse(x0y0,width=2*a1[i],height=2*b1[i],angle=theta[i]))
               e_sec.append(Ellipse(x0y0,width=2*a1_sec[i],height=2*b1_sec[i],angle=theta_sec[i]))

          #print 'a1_sec ',a1_sec
          #print 'b1_sec ',b1_sec
          #print 'theta_sec ',theta_sec
          #print 'mat_diag_sec ',mat_diag_sec




          rsample='latticestar'
          oxmax=XMax/self.lattice_calculator.modvec(o1[0],o1[1],o1[2],rsample)
          oxmin=XMin/self.lattice_calculator.modvec(o1[0],o1[1],o1[2],rsample)
          oymax=YMax/self.lattice_calculator.modvec(o2[0],o2[1],o2[2],rsample)
          oymin=YMin/self.lattice_calculator.modvec(o2[0],o2[1],o2[2],rsample)
          #print 'a1 ',a1
          #print 'b1 ',b1
          #print 'theta ',theta
          #print 'mat_diag ',mat_diag
          #x0y0=N.array([1.0,0.0])
          #make right y-axis
          ax2 = fig.add_subplot(2,2,1)
          pylab.subplots_adjust(hspace=0.6,wspace=0.3)
          #ax2.set_xlim(oxmin, oxmax)
          ax2.set_ylim(oymin[center], oymax[center])
          ax2.yaxis.tick_right()
          ax2.yaxis.set_label_position('right')
          ax2.xaxis.set_major_formatter(pylab.NullFormatter())
          ax2.xaxis.set_major_locator(pylab.NullLocator())
          ylabel=r'Q$_y$' +'(units of ['+str(o2[0,center])+' '+str(o2[1,center])+' '+str(o2[2,center])+'])'
          ax2.set_ylabel(ylabel)
          #ax2.set_zorder(3)
          #make top x-axis
          if 1:
               ax3 = fig.add_axes(ax2.get_position(), frameon=False,label='x-y top')
               ax3.xaxis.tick_top()
               ax3.xaxis.set_label_position('top')
               ax3.set_xlim(oxmin[center], oxmax[center])
               ax3.yaxis.set_major_formatter(NullFormatter())
               ax3.yaxis.set_major_locator(pylab.NullLocator())
               xlabel=r'Q$_x$' +'(units of ['+str(o1[0,center])+' '+str(o1[1,center])+' '+str(o1[2,center])+'])'
               ax3.set_xlabel(xlabel)
               #ax3.set_zorder(2)

          #make bottom x-axis, left y-axis
          if 1:
               ax = fig.add_axes(ax2.get_position(), frameon=False,label='x-y')
               ax.yaxis.tick_left()
               ax.yaxis.set_label_position('left')
               ax.xaxis.tick_bottom()
               #ax.xaxis.set_label_position('bottom')
               for i in range(c):
                    ax.add_artist(e[i])
                    e[i].set_clip_box(ax.bbox)
                    e[i].set_alpha(0.5)
                    e[i].set_facecolor('red')
                    ax.add_artist(e_sec[i])
                    e_sec[i].set_clip_box(ax.bbox)
                    e_sec[i].set_alpha(0.7)
                    e_sec[i].set_facecolor('blue')

               ax.set_xlim(XMin, XMax)
               ax.set_ylim(YMin, YMax)
               xlabel=r'Q$_x$ ('+r'$\AA^{-1}$)'
               ax.set_xlabel(xlabel)
               ylabel=r'Q$_y$ ('+r'$\AA^{-1}$)'
               ax.set_ylabel(ylabel)
               #ax.set_zorder(1)

          #%========================================================================================================
          #% plot XE projection


          proj,sec=self.project(RMS,1);
          (a,b,c)=N.shape(proj)
          mat=N.copy(proj)
          #print 'proj ', proj
          a1=[];b1=[];theta=[];a1_sec=[];b1_sec=[];theta_sec=[];e=[]; e_sec=[]
          for i in range(c):
               matm=N.matrix(mat[:,:,i])
               w,v=N.linalg.eig(matm)
               vm=N.matrix(v)
               vmt=vm.T
               mat_diag=vmt*matm*vm
               a1.append(1.0/N.sqrt(mat_diag[0,0]))
               b1.append(1.0/N.sqrt(mat_diag[1,1]))
               #thetar=N.arccos(vm[0,0])*N.sign(vm[0,0])
               thetar=-N.arctan2(vm[0,1],vm[0,0])
               theta.append(math.degrees(thetar))
          mat_sec=N.copy(sec)
          #print 'proj ', proj
          (a,b,c)=N.shape(sec)
          for i in range(c):
               matm_sec=N.matrix(mat_sec[:,:,i])
               w_sec,v_sec=N.linalg.eig(matm_sec)
               vm_sec=N.matrix(v)
               vmt_sec=vm_sec.T
               mat_diag_sec=vmt_sec*matm_sec*vm_sec
               a1_sec.append(1.0/N.sqrt(mat_diag_sec[0,0]))
               b1_sec.append(1.0/N.sqrt(mat_diag_sec[1,1]))
               #thetar_sec=N.arccos(vm_sec[0,0])*N.sign(vm_sec[0,0])
               thetar_sec=-N.arctan2(vm_sec[0,1],vm_sec[0,0])
               theta_sec.append(math.degrees(thetar_sec))
               x0y0=N.array([qx[i],qw[i]])
               e.append(Ellipse(x0y0,width=2*a1[i],height=2*b1[i],angle=theta[i]))
               e_sec.append(Ellipse(x0y0,width=2*a1_sec[i],height=2*b1_sec[i],angle=theta_sec[i]))
          rsample='latticestar'
          oxmax=XMax/self.lattice_calculator.modvec(o1[0],o1[1],o1[2],rsample)
          oxmin=XMin/self.lattice_calculator.modvec(o1[0],o1[1],o1[2],rsample)
          oymax=WMax
          oymin=WMin
          #print 'a1 ',a1
          #print 'b1 ',b1
          #print 'theta ',theta
          #print 'mat_diag ',mat_diag
          #x0y0=N.array([1.0,0.0])
          #make right y-axis
          ax2 = fig.add_subplot(2,2,3)
          #ax2.set_xlim(oxmin, oxmax)
          ax2.set_ylim(oymin, oymax)
          ax2.yaxis.tick_right()
          ax2.yaxis.set_label_position('right')
          ax2.xaxis.set_major_formatter(pylab.NullFormatter())
          ax2.xaxis.set_major_locator(pylab.NullLocator())
          ax2.yaxis.set_major_locator(pylab.NullLocator())
          ylabel=r'E' +'(meV)'
          #ax2.set_ylabel(ylabel)
          #ax2.set_zorder(3)
          #make top x-axis
          if 1:
               ax3 = fig.add_axes(ax2.get_position(), frameon=False,label='x-E top')
               ax3.xaxis.tick_top()
               ax3.xaxis.set_label_position('top')
               ax3.set_xlim(oxmin[center], oxmax[center])
               ax3.yaxis.set_major_formatter(NullFormatter())
               ax3.yaxis.set_major_locator(pylab.NullLocator())
               xlabel=r'Q$_x$' +'(units of ['+str(o1[0,center])+' '+str(o1[1,center])+' '+str(o1[2,center])+'])'
               ax3.set_xlabel(xlabel)
               #ax3.set_zorder(2)

          #make bottom x-axis, left y-axis
          if 1:
               ax = fig.add_axes(ax2.get_position(), frameon=False,label='x-E')
               ax.yaxis.tick_left()
               ax.yaxis.set_label_position('left')
               ax.xaxis.tick_bottom()
               #ax.xaxis.set_label_position('bottom')

               for i in range(c):
                    ax.add_artist(e[i])
                    e[i].set_clip_box(ax.bbox)
                    e[i].set_alpha(0.5)
                    e[i].set_facecolor('red')
                    ax.add_artist(e_sec[i])
                    e_sec[i].set_clip_box(ax.bbox)
                    e_sec[i].set_alpha(0.7)
                    e_sec[i].set_facecolor('blue')
               ax.set_xlim(XMin, XMax)
               ax.set_ylim(WMin, WMax)
               xlabel=r'Q$_x$ ('+r'$\AA^{-1}$)'
               ax.set_xlabel(xlabel)
               ylabel=r'E (meV)'
               ax.set_ylabel(ylabel)
               #ax.yaxis.set_major_formatter(NullFormatter())
               #ax.yaxis.set_major_locator(pylab.NullLocator())

               #ax.set_zorder(1)
          #pylab.show()
          #%========================================================================================================
          #% plot YE projection


          proj,sec=self.project(RMS,0);
          (a,b,c)=N.shape(proj)
          mat=N.copy(proj)
          #print 'proj ', proj
          a1=[];b1=[];theta=[];a1_sec=[];b1_sec=[];theta_sec=[];e=[]; e_sec=[]
          for i in range(c):
               matm=N.matrix(mat[:,:,i])
               w,v=N.linalg.eig(matm)
               vm=N.matrix(v)
               vmt=vm.T
               mat_diag=vmt*matm*vm
               a1.append(1.0/N.sqrt(mat_diag[0,0]))
               b1.append(1.0/N.sqrt(mat_diag[1,1]))
               #thetar=N.arccos(vm[0,0])*N.sign(vm[0,0])
               thetar=-N.arctan2(vm[0,1],vm[0,0])
               theta.append(math.degrees(thetar))
          #print 'proj ', proj
          (a,b,c)=N.shape(sec)
          mat_sec=N.copy(sec)
          for i in range(c):
               matm_sec=N.matrix(mat_sec[:,:,i])
               w_sec,v_sec=N.linalg.eig(matm_sec)
               vm_sec=N.matrix(v)
               vmt_sec=vm_sec.T
               mat_diag_sec=vmt_sec*matm_sec*vm_sec
               a1_sec.append(1.0/N.sqrt(mat_diag_sec[0,0]))
               b1_sec.append(1.0/N.sqrt(mat_diag_sec[1,1]))
               #thetar_sec=N.arccos(vm_sec[0,0])*N.sign(vm_sec[0,0])
               thetar_sec=-N.arctan2(vm_sec[0,1],vm_sec[0,0])
               theta_sec.append(math.degrees(thetar_sec))
               x0y0=N.array([qy[i],qw[i]])
               e.append(Ellipse(x0y0,width=2*a1[i],height=2*b1[i],angle=theta[i]))
               e_sec.append(Ellipse(x0y0,width=2*a1_sec[i],height=2*b1_sec[i],angle=theta_sec[i]))

          print 'qx ',qx
          print 'qw ',qw
          print 'theta_sec ',theta_sec
          rsample='latticestar'
          oxmax=YMax/self.lattice_calculator.modvec(o2[0],o2[1],o2[2],rsample)
          oxmin=YMin/self.lattice_calculator.modvec(o2[0],o2[1],o2[2],rsample)
          oymax=WMax
          oymin=WMin
          #print 'a1 ',a1
          #print 'b1 ',b1
          #print 'theta ',theta
          #print 'mat_diag ',mat_diag
          #x0y0=N.array([1.0,0.0])
          #make right y-axis
          ax2 = fig.add_subplot(2,2,4)
          #ax2.set_xlim(oxmin, oxmax)
          ax2.set_ylim(oymin, oymax)
          ax2.yaxis.tick_right()
          ax2.yaxis.set_label_position('right')
          ax2.xaxis.set_major_formatter(pylab.NullFormatter())
          ax2.xaxis.set_major_locator(pylab.NullLocator())
          ylabel=r'E' +'(meV)'
          ax2.set_ylabel(ylabel)
          #ax2.set_zorder(3)
          #make top x-axis
          if 1:
               ax3 = fig.add_axes(ax2.get_position(), frameon=False,label='y-E top')
               ax3.xaxis.tick_top()
               ax3.xaxis.set_label_position('top')
               ax3.set_xlim(oxmin[center], oxmax[center])
               ax3.yaxis.set_major_formatter(NullFormatter())
               ax3.yaxis.set_major_locator(pylab.NullLocator())
               xlabel=r'Q$_y$' +'(units of ['+str(o2[0,center])+' '+str(o2[1,center])+' '+str(o2[2,center])+'])'
               ax3.set_xlabel(xlabel)
               #ax3.set_zorder(2)

          #make bottom x-axis, left y-axis
          if 1:
               ax = fig.add_axes(ax2.get_position(), frameon=False,label='y-E')
               ax.yaxis.tick_left()
               ax.yaxis.set_label_position('left')
               ax.xaxis.tick_bottom()
               #ax.xaxis.set_label_position('bottom')

               for i in range(c):
                    ax.add_artist(e[i])
                    e[i].set_clip_box(ax.bbox)
                    e[i].set_alpha(0.5)
                    e[i].set_facecolor('red')
                    ax.add_artist(e_sec[i])
                    e_sec[i].set_clip_box(ax.bbox)
                    e_sec[i].set_alpha(0.7)
                    e_sec[i].set_facecolor('blue')

               ax.set_xlim(YMin, YMax)
               ax.set_ylim(WMin, WMax)
               xlabel=r'Q$_y$ ('+r'$\AA^{-1}$)'
               ax.set_xlabel(xlabel)
               #ylabel=r'E (meV)'
               #ax.set_ylabel(ylabel)
               #ax.yaxis.set_major_formatter(NullFormatter())
               ax.yaxis.set_major_locator(pylab.NullLocator())

               #ax.set_zorder(1)


          #pylab.show()
          pylab.savefig("resout.png")


          #self.PlotEllipse(proj,qx,qw,Style1);
          #self.PlotEllipse(sec,qx,qw,Style2);
          #pylab.show()
          return



     def fproject(self,mat_in,i):
          """return hwhm of projection"""
          if (i==0):
               v=2
               j=1
          if (i==1):
               v=0
               j=2
          if (i==2):
               v=0
               j=1
          mat=N.array(mat_in)
          (a,b,c)=N.shape(mat)
          proj=N.zeros((2,2,c),'d')
          proj[0,0,:]=mat[i,i,:]-mat[i,v,:]*mat[i,v,:]/mat[v,v,:]
          proj[0,1,:]=mat[i,j,:]-mat[i,v,:]*mat[j,v,:]/mat[v,v,:]
          proj[1,0,:]=mat[j,i,:]-mat[j,v,:]*mat[i,v,:]/mat[v,v,:]
          proj[1,1,:]=mat[j,j,:]-mat[j,v,:]*mat[j,v,:]/mat[v,v,:]
          hwhm=proj[0,0,:]-proj[0,1,:]*proj[0,1,:]/proj[1,1,:]
          hwhm=N.sqrt(2*N.log(2))/N.sqrt(hwhm)
          return hwhm


     def PlotEllipse(self,mat_in,x0,y0,style):
          """plot ellipse"""
          mat=N.array(mat_in)
          (a,b,c)=N.shape(mat)
               #phi=0:2*pi/3000:2*pi;
          phi=N.arange(0,2*pi,2*pi/3000)
          for i in range(c):
               r=N.sqrt(2*N.log(2)/(mat[0,0,i]*N.cos(phi)*N.cos(phi)+mat[1,1,i]*N.sin(phi)*N.sin(phi)\
                                    +2*mat[0,1,i]*N.cos(phi)*N.sin(phi)))
               x=r*N.cos(phi)+x0[i];
               y=r*N.sin(phi)+y0[i];
               pylab.plot(x,y,style);
          return

     def project(self,mat_in,v):
          """return projection and cross section matrices"""
          if v == 2:
               i=0;j=1
          if v == 0:
               i=1;j=2
          if v == 1:
               i=0;j=2
          mat=N.array(mat_in)
          (a,b,c)=N.shape(mat)
          proj=N.zeros((2,2,c),'d')
          sec=N.zeros((2,2,c),'d')
          proj[0,0,:]=mat[i,i,:]-mat[i,v,:]*mat[i,v,:]/mat[v,v,:]
          proj[0,1,:]=mat[i,j,:]-mat[i,v,:]*mat[j,v,:]/mat[v,v,:]
          proj[1,0,:]=mat[j,i,:]-mat[j,v,:]*mat[i,v,:]/mat[v,v,:]
          proj[1,1,:]=mat[j,j,:]-mat[j,v,:]*mat[j,v,:]/mat[v,v,:]
          sec[0,0,:]=mat[i,i,:]
          sec[0,1,:]=mat[i,j,:]
          sec[1,0,:]=mat[j,i,:]
          sec[1,1,:]=mat[j,j,:]
          return proj,sec

     def cooper_nathans_crystal_transform(self,H,K,L):
          "Provides a transformation matrix from the CooperNathan's Coordinate  system to the CrystalCoordinate system"
          x=self.lattice_calculator.x
          y=self.lattice_calculator.y
          z=self.lattice_calculator.z
          Q=self.lattice_calculator.modvec(H,K,L,'latticestar')
          uq=N.zeros((3,self.lattice_calculator.npts),'Float64')
          uq[0,:]=H/Q;  #% Unit vector along Q
          uq[1,:]=K/Q;
          uq[2,:]=L/Q;
          xq=self.lattice_calculator.scalar(x[0,:],x[1,:],x[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
          yq=self.lattice_calculator.scalar(y[0,:],y[1,:],y[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
          zq=0; # %scattering vector assumed to be in (self.orient1,self.orient2) plane;
          tmat=N.zeros((4,4,self.lattice_calculator.npts)); #%Coordinate transformation matrix
          tmat[3,3,:]=1;
          tmat[2,2,:]=1;
          tmat[0,0,:]=xq;
          tmat[0,1,:]=yq;
          tmat[1,1,:]=xq;
          tmat[1,0,:]=-yq;
          return tmat


     def calc_correction(self,H,K,L,W,EXP,qscan=None):
          "Returns the correction factor such that I_int=F^2*correction"
          Q=self.lattice_calculator.modvec(H,K,L,'latticestar')
          R0,RM=self.ResMat(Q,W,EXP)
          th_correction=[]
          tth_correction=[]
          if qscan != None:
               qx=[]
               qy=[]
               qmod=[]
               for i in range(len(qscan)):
                    qh=N.array([qscan[i][0]],'Float64')
                    qk=N.array([qscan[i][1]],'Float64')
                    ql=N.array([qscan[i][2]],'Float64')
                    tmat=(self.cooper_nathans_crystal_transform(qh,qk,ql))
                    qcooper=N.matrix(tmat[0:3,0:3,i])*N.matrix(qscan[i],'Float64').T
                    qscan_mod=self.lattice_calculator.modvec(qh,qk,ql,'latticestar')
                    #print 'qcooper ',qcooper[0,0]
                    qx.append(qcooper[0,0]/qscan_mod)
                    qy.append(qcooper[1,0]/qscan_mod)

          q_correction=[]
          #print 'qx ',qx[0][0]
          for i in range(RM.shape[2]):
               Myy=RM[1,1,i]
               Mxx=RM[0,0,i]
               Mxy=RM[0,1,i]
               th_correction.append(N.sqrt(N.linalg.det(RM[:,:,i]))/N.sqrt(Myy)/Q[i])

               #print 'RM',RM[:,:,i]
               #print 'det', N.sqrt(N.linalg.det(RM[:,:,i]))
               #print 'sqrt',N.sqrt(Myy)
               #print 'Q', Q[i]
               tth_correction.append(N.sqrt(N.linalg.det(RM[:,:,i]))/N.sqrt(Mxx)/Q[i])
               if qscan!=None:
                    q_correction.append(N.sqrt(N.linalg.det(RM[:,:,i]))/N.sqrt(Mxx*qx[i]**2+2*Mxy*qx[i]*qy[i]+Myy*qy[i]**2)/Q[i])
          corrections={}
          corrections['th_correction']=th_correction
          corrections['tth_correction']=tth_correction
          if qscan!=None:
               corrections['q_correction']=q_correction
          return corrections

     def CalcWidths(self,H,K,L,W,EXP):
          """Calculates the full width of Bragg Peaks"""
          Q=self.lattice_calculator.modvec(H,K,L,'latticestar')
          #print 'Q ',Q
          npts=self.lattice_calculator.npts
          #print 'npts ',npts
          center=N.round(H.shape[0]/2)
          if center<1:
               center=0
          if center>H.shape[0]:
               center=H.shape[0]
          #EXP=[EXP[center]]

          [R0c,RMc]=self.ResMat(Q,W,EXP)
          [R0,RMS]=self.ResMatS(H,K,L,W,EXP)
          self.lattice_calculator.StandardSystem()
          qx=self.lattice_calculator.scalar(self.lattice_calculator.x[0,:],self.lattice_calculator.x[1,:],self.lattice_calculator.x[2,:],H,K,L,'latticestar')
          qy=self.lattice_calculator.scalar(self.lattice_calculator.y[0,:],self.lattice_calculator.y[1,:],self.lattice_calculator.y[2,:],H,K,L,'latticestar')
          qw=W;

          #========================================================================================================
          #find reciprocal-space directions of X and Y axes

          o1=self.lattice_calculator.orientation.orient1.T #EXP['orient1']
          o2=self.lattice_calculator.orientation.orient2.T #EXP['orient2']
#        print 'o2shape ',o2.shape
#        print 'yshape ',self.lattice_calculator.y.shape
          pr=self.lattice_calculator.scalar(o2[0,:],o2[1,:],o2[2,:],self.lattice_calculator.y[0,:],self.lattice_calculator.y[1,:],self.lattice_calculator.y[2,:],'latticestar')
          o2[0]=self.lattice_calculator.y[0,:]*pr
          o2[1]=self.lattice_calculator.y[1,:]*pr
          o2[2]=self.lattice_calculator.y[2,:]*pr


          XWidth=max(self.fproject(RMS,0))
          YWidth=max(self.fproject(RMS,1))
          WWidth=max(self.fproject(RMS,2))

          proj,sec=self.project(RMS,1);
          (a,b,c)=N.shape(proj)
          mat=N.copy(proj)
          for i in range(c):
               matm=N.matrix(mat[:,:,i])
               w,v=N.linalg.eig(matm)
               vm=N.matrix(v)
               vmt=vm.T
               mat_diag=vmt*matm*vm
          a1=1.0/N.sqrt(mat_diag[0,0])
          b1=1.0/N.sqrt(mat_diag[1,1])
          thetar=N.arccos(vm[0,0])*N.sign(vm[0,0])
          theta=math.degrees(thetar)
          XWidth=2*self.fproject(RMS,0)
          YWidth=2*self.fproject(RMS,1)
          WWidth=2*self.fproject(RMS,2)
          ZWidth=2*N.sqrt(2*N.log(2))/N.sqrt(RMS[3,3,:])
          XBWidth=2*N.sqrt(2*N.log(2))/N.sqrt(RMS[0,0,:])
          YBWidth=2*N.sqrt(2*N.log(2))/N.sqrt(RMS[1,1,:])
          WBWidth=2*N.sqrt(2*N.log(2))/N.sqrt(RMS[2,2,:])

          XWidthc=1*self.fproject(RMc,0)
          YWidthc=1*self.fproject(RMc,1)
          ZWidthc=1*N.sqrt(2*N.log(2))/N.sqrt(RMc[3,3,:])
          XBWidthc=1*N.sqrt(2*N.log(2))/N.sqrt(RMc[0,0,:])
          YBWidthc=1*N.sqrt(2*N.log(2))/N.sqrt(RMc[1,1,:])

          CONVERT1=0.4246609*N.pi/60/180;
          CONVERT2=2.072;
          for ind in range(npts):
               infin=-1
               if 'infin' in EXP[ind]:
                    infin = EXP[ind]['infin']
               efixed=EXP[ind]['efixed']
               w=W[ind]
               q=Q[ind]
               ei=efixed
               ef=efixed
               if infin>0:
                    ef=efixed-w
               else:
                    ei=efixed+w;
               ki = N.sqrt(ei/CONVERT2)
               kf = N.sqrt(ef/CONVERT2)
          S2=N.arccos((ki**2+kf**2-Q**2)/(2*ki*kf))
                    #			todeg=1./reselpsq/!dtor
                    #			; Q^2 = ki^2 + kf^2 + 2 ki kf cos(A4)
                    #			; 2 Q dQ = 2 ki kf (-)sin(a4) d A4
                    #			; d a4 = -Q dQ / ( ki kf sin(A4) )
                    #			todega4=(reselpsq/ki/kf/sin(a4*!dtor))/!dtor

          tthwidth_sec=XBWidthc*2*Q/ki/kf/N.sin(S2)
          tthwidth_proj=XWidthc*2*Q/ki/kf/N.sin(S2)
          thwidth_sec=(YBWidthc*2)/Q  # recall for small theta, arc=r*theta
          thwidth_proj=(YWidthc*2)/Q


##        #M1,M2,S1,S2,A1,A2=mylattice.SpecGoTo(H+XWidth,K,L,W,EXP)
##        uq=N.zeros((3,self.lattice_calculator.npts),'Float64')
##        #Ql=N.sqrt(H**2+K**2+L**2)
##        uq[0,:]=H/Q  #% Unit vector along Q
##        uq[1,:]=K/Q
##        uq[2,:]=L/Q
##        #projection
##        Hup=uq[0,:]*XWidthc
##        Kup=uq[1,:]*XWidthc
##        Lup=uq[2,:]*XWidthc
##        M1,M2,S1,S2,A1,A2=mylattice.SpecGoTo(H+Hup,K+Kup,L+Lup,W,EXP)
##        M1c,M2c,S1c,S2c,A1c,A2c=mylattice.SpecGoTo(H,K,L,W,EXP)
##        tthwidth_proj=N.absolute(S2-S2c)*2
##        #cross section
##        Hup=uq[0,:]*XBWidthc
##        Kup=uq[1,:]*XBWidthc
##        Lup=uq[2,:]*XBWidthc
##        print 'Hup ', Hup
##        print 'Hadded ', H+Hup
##        print 'Kup ', Kup
##        print 'Kadded ', K+Kup
##        print 'Lup ', Lup
##        print 'Ladded ', L+Lup
##
##        M1,M2,S1,S2,A1,A2=self.lattice_calculator.SpecGoTo(H+Hup,K+Kup,L+Lup,W,EXP)
##        M1c,M2c,S1c,S2c,A1c,A2c=self.lattice_calculator.SpecGoTo(H,K,L,W,EXP)
##        tthwidth_sec=N.absolute(S2-S2c)*2
##        print 'S2 ',math.degrees(S2)
##        print 'S2c ',math.degrees(S2c)
##        print 'tthwidth_sec ',tthwidth_sec

##        XWidth=fproject (RMS,1);
##        YWidth=fproject (RMS,2);
##        WWidth=fproject (RMS,3);
##        ZWidth=sqrt(2*log(2))./sqrt(RMS(4,4,:));
##
##        XBWidth=sqrt(2*log(2))./sqrt(RMS(1,1,:));
##        YBWidth=sqrt(2*log(2))./sqrt(RMS(2,2,:));
##        WBWidth=sqrt(2*log(2))./sqrt(RMS(3,3,:));
##
##
##        ResVol=(2*pi)^2/sqrt(det(RMS(:,:,center)));


          Widths={}
          Widths['XWidth']=XWidth
          Widths['YWidth']=YWidth
          Widths['ZWidth']=ZWidth
          Widths['WWidth']=WWidth
          Widths['XBWidth']=XBWidth
          Widths['YBWidth']=YBWidth
          Widths['WBWidth']=WBWidth
          Widths['tthwidth_proj']=N.degrees(tthwidth_proj)
          Widths['tthwidth_sec']=N.degrees(tthwidth_sec)
          Widths['thwidth_proj']=N.degrees(thwidth_proj)
          Widths['thwidth_sec']=N.degrees(thwidth_sec)
          return Widths


class TestLattice(unittest.TestCase):

     def setUp(self):
          a=N.array([2*pi],'d')
          b=N.array([2*pi],'d')
          c=N.array([2*pi],'d')
          alpha=N.array([90],'d')
          beta=N.array([90],'d')
          gamma=N.array([90],'d')
          orient1=N.array([[1,0,0]],'d')
          orient2=N.array([[0,1,1]],'d')
          self.fixture = lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                                 orient1=orient1,orient2=orient2)

     def test_astar(self):
          self.assertAlmostEqual(self.fixture.astar[0],1.0,2,'astar Not equal to '+str(1.0))
     def test_bstar(self):
          self.assertAlmostEqual(self.fixture.bstar[0],1.0,2,'bstar Not equal to '+str(1.0))
     def test_cstar(self):
          self.assertAlmostEqual(self.fixture.cstar[0],1.0,2,'cstar '+str(self.fixture.cstar[0])+' Not equal to '+str(1.0))
     def test_alphastar(self):
          self.assertAlmostEqual(self.fixture.alphastar[0],pi/2,2,'alphastar Not equal to '+str(pi/2))
     def test_betastar(self):
          self.assertAlmostEqual(self.fixture.betastar[0],pi/2,2,'betastar Not equal to '+str(pi/2))
     def test_gammastar(self):
          self.assertAlmostEqual(self.fixture.gammastar[0],pi/2,2,'gammastar Not equal to '+str(pi/2))
     def test_V(self):
          self.assertAlmostEqual(self.fixture.V[0],248.0502,2,'V Not equal to '+str(248.0502))
     def test_Vstar(self):
          self.assertAlmostEqual(self.fixture.Vstar[0],1.0,2,'Vstar Not equal to '+str(1.0))
     def test_g(self):
          #print self.fixture.g
          self.assertAlmostEqual((self.fixture.g[:,:,0][0,0]),39.4784*(N.eye(3)[0,0]) ,2,'g Not equal to '+str(39.4784 ))
     def test_gstar(self):
          #print self.fixture.gstar
          self.assertAlmostEqual(self.fixture.gstar[:,:,0][0,0],1.0*N.eye(3)[0,0] ,2,'gstar Not equal to '+str(1.0 ))

     def test_StandardSystem_x(self):
     #       #print self.fixture.gstar
          self.assertAlmostEqual(self.fixture.x[0],1.0 ,2,'Standard System x Not equal to '+str(1.0 ))




#    def test_zeroes(self):
#        self.assertEqual(0 + 0, 0)
#        self.assertEqual(5 + 0, 5)
#        self.assertEqual(0 + 13.2, 13.2)
#
#    def test_positive(self):
#        self.assertEqual(123 + 456, 579)
#        self.assertEqual(1.2e20 + 3.4e20, 3.5e20)
#
#    def test_mixed(self):
#        self.assertEqual(-19 + 20, 1)
#        self.assertEqual(999 + -1, 998)
#        self.assertEqual(-300.1 + -400.2, -700.3)
#


def get_tokenized_line(myfile,returnline=['']):
     lineStr=myfile.readline()
     returnline[0]=lineStr.rstrip()
     strippedLine=lineStr.lower().rstrip()
     tokenized=strippedLine.split()

     return tokenized

if __name__=="__main__":
     
     if 0:
          if 1:
               print 'trying'
               try:
                    infile=sys.argv[1]
               except:
                    infile=r'rescalc_in.txt'
               myfile=open(infile,'r')
               myFlag=True
               returnline=['']
               i=0
               EXP={}
               EXP['ana']={}
               EXP['mono']={}
               EXP['sample']={}
               EXP['method']=0
               while myFlag:
                    toks=get_tokenized_line(myfile,returnline=returnline)
                    #print 'toks',toks
                    if toks==[]:
                         break
                    if toks[0].startswith('#'):
                         continue
                    else:
                         if i==0:
                              a=N.array([float(toks[0])])
                              b=N.array([float(toks[1])])
                              c=N.array([float(toks[2])])
                              alpha=N.radians(N.array([float(toks[3])]))
                              beta=N.radians(N.array([float(toks[4])]))
                              gamma=N.radians(N.array([float(toks[5])]))
                         elif i==1:
                              EXP['hcol']=N.array([float(toks[0]),
                                                   float(toks[1]),
                                                   float(toks[2]),
                                                   float(toks[3])],'d')
                         elif i==2:
                              EXP['vcol']=N.array([float(toks[0]),
                                                   float(toks[1]),
                                                   float(toks[2]),
                                                   float(toks[3])],'d')
                         elif i==3:
                              EXP['mono']['mosaic']=float(toks[0])
                         elif i==4:
                              EXP['ana']['mosaic']=float(toks[0])
                         elif i==5:
                              EXP['sample']['mosaic']=float(toks[0])
                         elif i==6:
                              EXP['sample']['vmosaic']=float(toks[0])
                         elif i==7:
                              EXP['efixed']=float(toks[0])
                         elif i==8:
                              EXP['infix']=1 if toks[0]=='Ei' else -1
                         elif i==9:
                              EXP['mono']['tau']=str(toks[0])
                         elif i==10:
                              EXP['ana']['tau']=str(toks[0])
                         elif i==11:
                              orient1=N.array([toks],'d')
                         elif i==12:
                              orient2=N.array([toks],'d')
                         elif i==13:
                              H=N.array(toks,'d')
                         elif i==14:    
                              K=N.array(toks,'d')
                         elif i==15:     
                              L=N.array(toks,'d')
                         elif i==16:     
                              W=N.array(toks,'d')             
                         i=i+1
               orientation=lattice_calculator.Orientation(orient1,orient2)
               mylattice=lattice_calculator.Lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                                               orientation=orientation)
               setup=[EXP]
               myrescal=rescalculator(mylattice)
               newinput=lattice_calculator.CleanArgs(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,orient1=orient1,orient2=orient2,\
                                                     H=H,K=K,L=L,W=W,setup=setup)
               neworientation=lattice_calculator.Orientation(newinput['orient1'],newinput['orient2'])
               mylattice=lattice_calculator.Lattice(a=newinput['a'],b=newinput['b'],c=newinput['c'],alpha=newinput['alpha'],\
                                                    beta=newinput['beta'],gamma=newinput['gamma'],orientation=neworientation)
               myrescal.__init__(mylattice)
               Q=myrescal.lattice_calculator.modvec(H,K,L,'latticestar')
               print 'Q', Q
               #sys.exit()
               R0,RM=myrescal.ResMat(Q,W,setup)
               print 'RM '
               print RM.transpose()
               print 'R0 ',R0
               #exit()
               #print "shapes"
               #print H.shape
               #print K.shape
               #print L.shape
               #print W.shape
               R0,RMS=myrescal.ResMatS(H,K,L,W,setup)
               myrescal.ResPlot(H, K, L, W, setup)
               print 'RMS'
               print RMS.transpose()[0]
               #print myrescal.calc_correction(H,K,L,W,setup,qscan=[[1,0,1],[1,0,1]])
               print myrescal.calc_correction(H,K,L,W,setup)
               print myrescal.CalcWidths(H,K,L,W,setup)
               print 'setup length ',len(setup)
               sys.exit()
          if 0:
               sys.exit()
          
     if 0:
          a=N.array([2*pi],'d')
          b=N.array([2*pi],'d')
          c=N.array([2*pi],'d')
          alpha=N.array([90],'d')
          beta=N.array([90],'d')
          gamma=N.array([90],'d')
     #       orient1=N.array([[0,1,1]],'d')
          orient1=N.array([[1,0,0]],'d')
          orient2=N.array([[0,1,1]],'d')
          mylattice=lattice_calculator.lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                                               orient1=orient1,orient2=orient2)
          H=N.array([1],'d');K=N.array([0],'d');L=N.array([0],'d');W=N.array([0],'d')
          EXP={}
          EXP['ana']={}
          EXP['ana']['tau']='pg(002)'
          EXP['mono']={}
          EXP['mono']['tau']='pg(002)';
          EXP['ana']['mosaic']=30
          EXP['mono']['mosaic']=30
          EXP['sample']={}
          EXP['sample']['mosaic']=10
          EXP['sample']['vmosaic']=10
          EXP['hcol']=N.array([40, 10, 20, 80],'d')
          EXP['vcol']=N.array([120, 120, 120, 120],'d')
          EXP['infix']=-1 #positive for fixed incident energy
          EXP['efixed']=14.7
          EXP['method']=0
          setup=[EXP]
          myrescal=rescalculator(mylattice)
          #R0,RMS=myrescal.ResMatS(H,K,L,W,setup)
          R0,RM= self.ResMat(Q,W,setup)
          #myrescal.ResPlot(H, K, L, W, setup)
          print 'RMS'
          print RMS.transpose()[0]
          print myrescal.calc_correction(H,K,L,W,setup)
#        mylattice.StandardSystem()
##    x1=N.array([1.0, 1.0], 'd'); y1=N.array([1.0, 1.0], 'd'); z1=N.array([1.0, 1.0], 'd'); x2=x1; y2=y1; z2=z1;
##    print 'scalar ', mylattice.scalar(x1,y1,z1,x2,y2,z2,'lattice')
##    print 'me ',mylattice.gtensor('lattice')
##    myinstrument=instrument();
##    print myinstrument.get_tau('pg(004)')
#    unittest.main()


     if 1:
          a=N.array([6, 6.],'d')
          b=N.array([7., 7.],'d')
          c=N.array([8.,8],'d')
          alpha=N.array([pi/2,pi/2],'d')
          beta=N.array([pi/2,pi/2],'d')
          gamma=N.array([pi/2,pi/2],'d')
     #       orient1=N.array([[0,1,1]],'d')
          orient1=N.array([[1,0,0],[1,0,0]],'d')
          orient2=N.array([[0,0,1],[0,0,1]],'d')
          orientation=lattice_calculator.Orientation(orient1,orient2)
          mylattice=lattice_calculator.Lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                                               orientation=orientation)
          H=N.array([1,1],'d');K=N.array([0,0.0],'d');L=N.array([5.0,5.0],'d');W=N.array([2,2],'d')
          EXP={}
          EXP['ana']={}
          EXP['ana']['tau']='pg(002)'
          EXP['mono']={}
          EXP['mono']['tau']='pg(002)';
          EXP['ana']['mosaic']=25
          EXP['mono']['mosaic']=25
          EXP['sample']={}
          EXP['sample']['mosaic']=25
          EXP['sample']['vmosaic']=25
          EXP['hcol']=N.array([40, 40, 40, 40],'d')
          EXP['vcol']=N.array([120, 120, 120, 120],'d')
          EXP['infix']=-1 #positive for fixed incident energy
          EXP['efixed']=14.7
          EXP['method']=0
          setup=[EXP]
          myrescal=rescalculator(mylattice)
          newinput=lattice_calculator.CleanArgs(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,orient1=orient1,orient2=orient2,\
                                                H=H,K=K,L=L,W=W,setup=setup)
          neworientation=lattice_calculator.Orientation(newinput['orient1'],newinput['orient2'])
          mylattice=lattice_calculator.Lattice(a=newinput['a'],b=newinput['b'],c=newinput['c'],alpha=newinput['alpha'],\
                                               beta=newinput['beta'],gamma=newinput['gamma'],orientation=neworientation)
          myrescal.__init__(mylattice)
          Q=myrescal.lattice_calculator.modvec(H,K,L,'latticestar')
          print 'Q', Q
          R0,RM=myrescal.ResMat(Q,W,setup)
          print 'RM '
          print RM.transpose()
          print 'R0 ',R0
          #exit()
          R0,RMS=myrescal.ResMatS(H,K,L,W,setup)
          myrescal.ResPlot(H, K, L, W, setup)
          print 'RMS'
          print RMS.transpose()[0]
          #print myrescal.calc_correction(H,K,L,W,setup,qscan=[[1,0,1],[1,0,1]])
          print myrescal.calc_correction(H,K,L,W,setup)
          print myrescal.CalcWidths(H,K,L,W,setup)
          print 'setup length ',len(setup)
          



"""
NumPy backend for CPU computations.
"""

import numpy as np
from scipy import linalg as scipy_linalg
from .base import BaseBackend


class NumpyBackend(BaseBackend):
    """NumPy-based backend for CPU computations."""

    def __init__(self):
        self._device = 'cpu'
        self.dtype = np.float64

    @property
    def name(self):
        return 'numpy'

    @property
    def device(self):
        return self._device

    def to_array(self, data, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return np.asarray(data, dtype=dtype)

    def to_numpy(self, data):
        return np.asarray(data)

    def zeros(self, shape, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return np.zeros(shape, dtype=dtype)

    def ones(self, shape, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return np.ones(shape, dtype=dtype)

    def eye(self, n, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return np.eye(n, dtype=dtype)

    def diag(self, v):
        return np.diag(v)

    def dot(self, a, b):
        return np.dot(a, b)

    def matmul(self, a, b):
        return np.matmul(a, b)

    def inv(self, a):
        return np.linalg.inv(a)

    def det(self, a):
        return np.linalg.det(a)

    def eig(self, a):
        return np.linalg.eig(a)

    def sqrt(self, x):
        return np.sqrt(x)

    def exp(self, x):
        return np.exp(x)

    def sin(self, x):
        return np.sin(x)

    def cos(self, x):
        return np.cos(x)

    def tan(self, x):
        return np.tan(x)

    def arcsin(self, x):
        return np.arcsin(x)

    def arccos(self, x):
        return np.arccos(x)

    def arctan2(self, y, x):
        return np.arctan2(y, x)

    def abs(self, x):
        return np.abs(x)

    def log(self, x):
        return np.log(x)

    def sum(self, x, axis=None):
        return np.sum(x, axis=axis)

    def stack(self, arrays, axis=0):
        return np.stack(arrays, axis=axis)

    def concatenate(self, arrays, axis=0):
        return np.concatenate(arrays, axis=axis)

    def transpose(self, a, axes=None):
        return np.transpose(a, axes)

    def reshape(self, a, shape):
        return np.reshape(a, shape)

    def linspace(self, start, stop, num):
        return np.linspace(start, stop, num)

    def arange(self, start, stop=None, step=1):
        if stop is None:
            return np.arange(start)
        return np.arange(start, stop, step)

    def meshgrid(self, *xi, indexing='xy'):
        return np.meshgrid(*xi, indexing=indexing)

    def einsum(self, subscripts, *operands):
        return np.einsum(subscripts, *operands)

    def block_diag(self, *matrices):
        return scipy_linalg.block_diag(*matrices)

    # Additional numpy-specific methods
    def sign(self, x):
        return np.sign(x)

    def size(self, x):
        return np.size(x)

    def copy(self, x):
        return np.copy(x)

    def clip(self, x, a_min, a_max):
        return np.clip(x, a_min, a_max)

    def where(self, condition, x, y):
        return np.where(condition, x, y)

    def maximum(self, x, y):
        return np.maximum(x, y)

    def minimum(self, x, y):
        return np.minimum(x, y)



import numpy as N

#def mytile(h,k):
#    """takes input arrays h,k and returns a grid""


if __name__=='__main__':
    ht=N.linspace(-1,1,21)
    kt=N.linspace(-2,2,41)
    h,k=N.meshgrid(ht,kt)
    print h.flatten(),k.flatten()